"""Tests for the document stores."""

import pytest
from scholarag.document_stores.elastic import postprocess_query


# Sync functions
def test_create_and_remove_index(get_testing_ds_client):
    """Test the creation and removal of an index."""
    ds_client, parameters = get_testing_ds_client

    index = "test_index"

    ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])

    assert index in ds_client.get_available_indexes()

    ds_client.remove_index(index)

    assert index not in ds_client.get_available_indexes()

    # Errors
    with pytest.raises(RuntimeError):
        ds_client.remove_index("does-not-exists")


def test_create_index_already_exists(get_testing_ds_client):
    """Test create index that already exists."""
    ds_client, parameters = get_testing_ds_client
    index = "test_index"
    ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])

    with pytest.raises(RuntimeError):
        ds_client.create_index(
            "test_index", settings=parameters[-1], mappings=parameters[0]
        )


def test_remove_index_error(get_testing_ds_client):
    """Test that we have runtime error when index does not exist."""
    ds_client, _ = get_testing_ds_client
    with pytest.raises(RuntimeError):
        ds_client.remove_index("does-not-exist")


def test_get_available_indexes(get_testing_ds_client):
    """Test the retrieval of available indexes."""
    ds_client, parameters = get_testing_ds_client

    index = "test_index"

    ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])

    assert "test_index" in ds_client.get_available_indexes()


def test_get_index_mappings(get_testing_ds_client):
    """Test the retrieval of the mapping of an index."""
    ds_client, parameters = get_testing_ds_client

    index = "test_index"

    ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])

    assert ds_client.get_index_mappings(index) == parameters[0]


def test_add_fields(get_testing_ds_client):
    ds_client, parameters = get_testing_ds_client

    index = "test_index"

    ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])
    doc = {"title": "test", "text": "test"}
    ds_client.add_document(index, doc)
    ds_client.client.indices.refresh()
    properties = {"test_field": {"type": "keyword"}}
    ds_client.add_fields(index=index, mapping=properties)
    mappings = ds_client.get_index_mappings(index)
    expected_mapping = parameters[0]
    expected_mapping["properties"]["test_field"] = {"type": "keyword"}
    assert mappings == expected_mapping

    # Error
    ds_client, _ = get_testing_ds_client
    with pytest.raises(RuntimeError):
        ds_client.add_fields("does-not-exist")


def test_count_documents(get_testing_ds_client):
    """Test the retrieval of the number of documents in an index."""
    ds_client, parameters = get_testing_ds_client

    index = "test_index"

    ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])

    assert ds_client.count_documents(index) == 0

    doc = {"title": "test", "text": "test"}
    ds_client.add_document(index, doc)
    ds_client.client.indices.refresh()

    assert ds_client.count_documents(index) == 1

    # Error
    ds_client, _ = get_testing_ds_client
    with pytest.raises(RuntimeError):
        ds_client.count_documents("does-not-exist")


def test_iter_document(get_testing_ds_client):
    """Test the retrieval of the number of documents in an index."""
    ds_client, parameters = get_testing_ds_client

    index = "test_index"

    ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])

    assert ds_client.count_documents(index) == 0

    doc = {"title": "test", "text": "test"}
    ds_client.add_document(index, doc, doc_id="doc_id1")
    ds_client.client.indices.refresh()

    gen = ds_client.iter_document(index=index)

    assert len(list(gen)) == 1

    # Errors
    with pytest.raises(RuntimeError):
        # Index does not exist
        ds_client.add_document("does-not-exist", doc, doc_id="doc_id1")

    with pytest.raises(RuntimeError):
        # Doc already exists
        ds_client.add_document(index, doc, doc_id="doc_id1")


def test_exists(get_testing_ds_client):
    """Test the retrieval of the number of documents in an index."""
    ds_client, parameters = get_testing_ds_client

    index = "test_index"

    ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])
    doc_id = 1
    doc = {"title": "test", "text": "test"}

    assert not ds_client.exists(index, doc_id)

    ds_client.add_document(index, doc, doc_id)
    ds_client.client.indices.refresh()

    assert ds_client.exists(index, doc_id)


def test_get_document(get_testing_ds_client):
    """Test the retrieval of a document."""
    ds_client, parameters = get_testing_ds_client

    index = "test_index"

    ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])

    doc = {"title": "test", "text": "test"}

    ds_client.add_document(index, doc, doc_id="test_id")
    ds_client.client.indices.refresh()

    assert ds_client.get_document(index, "test_id") == doc


def test_get_document_errors(get_testing_ds_client):
    """Test the errors when retrieval of a document."""
    ds_client, parameters = get_testing_ds_client
    with pytest.raises(RuntimeError):
        ds_client.get_document("does-not-exists", "doc-id-1")

    ds_client.create_index(
        "test_index", settings=parameters[-1], mappings=parameters[0]
    )
    with pytest.raises(RuntimeError):
        ds_client.get_document("test_index", "does-not-exist")


def test_get_documents(get_testing_ds_client):
    """Test the retrieval of a document."""
    ds_client, parameters = get_testing_ds_client

    index = "test_index"

    ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])

    doc_1 = {"title": "test_1", "text": "text_1"}
    doc_2 = {"title": "test_2", "text": "text_2"}

    ds_client.add_document(index, doc_1, doc_id="1")
    ds_client.add_document(index, doc_2, doc_id="2")
    ds_client.client.indices.refresh()

    expected = [{"document_id": "1", **doc_1}, {"document_id": "2", **doc_2}]
    assert ds_client.get_documents(index, ["1", "2"]) == expected
    assert ds_client.get_documents(index, ["1", "3"]) == [expected[0]]


def test_get_documents_errors(get_testing_ds_client):
    ds_client, _ = get_testing_ds_client
    with pytest.raises(RuntimeError):
        ds_client.get_documents("does-not-exist", ["doc-id1", "doc-id2"])


def test_bulk(get_testing_ds_client):
    ds_client, parameters = get_testing_ds_client

    index = "test_index"

    ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])
    doc = [
        {
            "_index": index,
            "_id": i,
            "_source": {
                "article_id": i,
                "text": f"Amazing text {i}",
                "paragraph_id": i,
            },
        }
        for i in range(10)
    ]

    ds_client.bulk(doc)
    ds_client.client.indices.refresh()
    n_doc = ds_client.count_documents(index=index)
    assert n_doc == 10
    retrieved_doc = ds_client.get_document(index=index, doc_id=3)["text"]
    assert retrieved_doc == "Amazing text 3"


@pytest.mark.parametrize("query", [{"match_all": {}}, {"match": {"text": "retrieve"}}])
def test_search(get_testing_ds_client, query):
    """Test the retrieval of a document."""
    ds_client, parameters = get_testing_ds_client

    index_doc = "test_paragraphs"
    aggs = {"unique_ids": {"terms": {"field": "paragraph_id", "size": 10}}}
    ds_client.create_index(index_doc, settings=parameters[-1], mappings=parameters[0])

    doc_1 = {
        "_index": index_doc,
        "_id": 1,
        "_source": {
            "text": "test of an amazing function",
            "title": "test_article",
            "paragraph_id": "1",
            "article_id": "article_id",
            "journal": "8765-4321",
        },
    }

    doc_2 = {
        "_index": index_doc,
        "_id": 2,
        "_source": {
            "text": "The bird sings very loudly",
            "title": "test_article",
            "paragraph_id": "1",
            "article_id": "article_id",
            "journal": None,
        },
    }

    doc_3 = {
        "_index": index_doc,
        "_id": 3,
        "_source": {
            "text": "This document is a bad test, I don't want to retrieve it",
            "title": "test_article",
            "paragraph_id": "1",
            "article_id": "article_id",
            "journal": "1234-5678",
        },
    }
    doc_bulk = [doc_1, doc_2, doc_3]

    ds_client.bulk(doc_bulk)

    ds_client.client.indices.refresh()

    results = ds_client.search(index_doc, query)
    res_hits = results["hits"]["hits"]
    assert res_hits[0]["_source"]["article_id"] == "article_id"
    assert res_hits[0]["_source"]["title"] == "test_article"
    if "match_all" in query.keys():
        assert len(res_hits) == 3
        assert res_hits[0]["_source"]["text"] == "test of an amazing function"
        assert res_hits[0]["_score"] == 1.0

        # test query + aggs.
        results = ds_client.search(index_doc, query, aggs=aggs)
        assert results["aggregations"]["unique_ids"]["buckets"][0]["doc_count"] == 3
    else:
        assert len(res_hits) == 1
        assert (
            res_hits[0]["_source"]["text"]
            == "This document is a bad test, I don't want to retrieve it"
        )
        assert res_hits[0]["_score"] == pytest.approx(0.7782316)
        # test query + aggs.
        results = ds_client.search(index_doc, query, aggs=aggs)
        assert results["aggregations"]["unique_ids"]["buckets"][0]["doc_count"] == 1

    # test aggs alone.
    results = ds_client.search(index_doc, aggs=aggs)
    assert results["aggregations"]["unique_ids"]["buckets"][0]["doc_count"] == 3

    # test without anything.
    results = ds_client.search(index_doc)
    assert "aggregations" not in results.keys()
    assert len(results["hits"]["hits"]) == 3


@pytest.mark.parametrize("filter_db", [None, {"match": {"text": "retrieve"}}])
def test_bm25_search(get_testing_ds_client, filter_db):
    """Test the retrieval of a document."""
    ds_client, parameters = get_testing_ds_client

    query = "test"
    index_doc = "test_paragraphs"

    ds_client.create_index(index_doc, settings=parameters[-1], mappings=parameters[0])

    doc_1 = {
        "_index": index_doc,
        "_id": 1,
        "_source": {
            "text": "test of an amazing function",
            "title": "test_article",
            "paragraph_id": "1",
            "article_id": "article_id",
            "journal": "8765-4321",
        },
    }

    doc_2 = {
        "_index": index_doc,
        "_id": 2,
        "_source": {
            "text": "The bird sings very loudly",
            "title": "test_article",
            "paragraph_id": "1",
            "article_id": "article_id",
            "journal": None,
        },
    }

    doc_3 = {
        "_index": index_doc,
        "_id": 3,
        "_source": {
            "text": "This document is a bad test, I don't want to retrieve it",
            "title": "test_article",
            "paragraph_id": "1",
            "article_id": "article_id",
            "journal": "1234-5678",
        },
    }
    doc_bulk = [doc_1, doc_2, doc_3]

    ds_client.bulk(doc_bulk)
    ds_client.client.indices.refresh()

    res = ds_client.bm25_search(index_doc, query, filter_db, 2)
    assert res[0]["article_id"] == "article_id"
    assert res[0]["title"] == "test_article"
    if filter_db is None:
        assert len(res) == 2
        assert res[0]["text"] == "test of an amazing function"
        assert res[0]["score"] == pytest.approx(0.5403367)
    else:
        assert len(res) == 1
        assert (
            res[0]["text"] == "This document is a bad test, I don't want to retrieve it"
        )
        assert res[0]["score"] == pytest.approx(0.37292093)

    # Errors
    with pytest.raises(RuntimeError):
        ds_client.bm25_search("does-not-exists", query, filter_db, 2)


# Async functions
@pytest.mark.asyncio
async def test_acreate_and_remove_index(get_testing_async_ds_client):
    """Test the creation and removal of an index."""
    ds_client, parameters = get_testing_async_ds_client

    index = "test_index"

    await ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])

    assert index in await ds_client.get_available_indexes()

    await ds_client.remove_index(index)

    assert index not in await ds_client.get_available_indexes()


@pytest.mark.asyncio
async def test_aget_available_indexes(get_testing_async_ds_client):
    """Test the retrieval of available indexes."""
    ds_client, parameters = get_testing_async_ds_client

    index = "test_index"

    await ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])

    assert "test_index" in await ds_client.get_available_indexes()


@pytest.mark.asyncio
async def test_aget_index_mappings(get_testing_async_ds_client):
    """Test the retrieval of the mapping of an index."""
    ds_client, parameters = get_testing_async_ds_client

    index = "test_index"

    await ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])

    assert await ds_client.get_index_mappings(index) == parameters[0]


@pytest.mark.asyncio
async def test_aadd_fields(get_testing_async_ds_client):
    ds_client, parameters = get_testing_async_ds_client

    index = "test_index"

    await ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])
    doc = {"title": "test", "text": "test"}
    await ds_client.add_document(index, doc)
    await ds_client.client.indices.refresh()
    properties = {"test_field": {"type": "keyword"}}
    await ds_client.add_fields(index=index, mapping=properties)
    mappings = await ds_client.get_index_mappings(index)
    expected_mapping = parameters[0]
    expected_mapping["properties"]["test_field"] = {"type": "keyword"}
    assert mappings == expected_mapping


@pytest.mark.asyncio
async def test_acount_documents(get_testing_async_ds_client):
    """Test the retrieval of the number of documents in an index."""
    ds_client, parameters = get_testing_async_ds_client

    index = "test_index"

    await ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])

    assert await ds_client.count_documents(index) == 0

    doc = {"title": "test", "text": "test"}
    await ds_client.add_document(index, doc)
    await ds_client.client.indices.refresh()

    assert await ds_client.count_documents(index) == 1


@pytest.mark.asyncio
async def test_aexists(get_testing_async_ds_client):
    """Test the retrieval of the number of documents in an index."""
    ds_client, parameters = get_testing_async_ds_client

    index = "test_index"

    await ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])
    doc_id = 1
    doc = {"title": "test", "text": "test"}

    assert not await ds_client.exists(index, doc_id)

    await ds_client.add_document(index, doc, doc_id)
    await ds_client.client.indices.refresh()

    assert await ds_client.exists(index, doc_id)


@pytest.mark.asyncio
async def test_aiter_document(get_testing_async_ds_client):
    """Test the retrieval of the number of documents in an index."""
    ds_client, parameters = get_testing_async_ds_client

    index = "test_index"

    await ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])

    assert await ds_client.count_documents(index) == 0

    doc = {"title": "test", "text": "test"}
    await ds_client.add_document(index, doc)
    await ds_client.client.indices.refresh()

    gen = ds_client.iter_document(index=index)
    gen = [item async for item in gen]

    assert len(list(gen)) == 1


@pytest.mark.asyncio
async def test_aget_document(get_testing_async_ds_client):
    """Test the retrieval of a document."""
    ds_client, parameters = get_testing_async_ds_client

    index = "test_index"

    await ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])

    doc = {"title": "test", "text": "test"}

    await ds_client.add_document(index, doc, doc_id="test_id")
    await ds_client.client.indices.refresh()

    assert await ds_client.get_document(index, "test_id") == doc

    with pytest.raises(RuntimeError):
        await ds_client.get_document(index, "does-not-exists")


@pytest.mark.asyncio
async def test_aget_documents(get_testing_async_ds_client):
    """Test the retrieval of a document."""
    ds_client, parameters = get_testing_async_ds_client

    index = "test_index"

    await ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])

    doc_1 = {"title": "test_1", "text": "text_1"}
    doc_2 = {"title": "test_2", "text": "text_2"}

    await ds_client.add_document(index, doc_1, doc_id="1")
    await ds_client.add_document(index, doc_2, doc_id="2")
    await ds_client.client.indices.refresh()

    expected = [{"document_id": "1", **doc_1}, {"document_id": "2", **doc_2}]
    assert await ds_client.get_documents(index, ["1", "2"]) == expected
    assert await ds_client.get_documents(index, ["1", "3"]) == [expected[0]]


@pytest.mark.asyncio
async def test_abulk(get_testing_async_ds_client):
    ds_client, parameters = get_testing_async_ds_client

    index = "test_index"

    await ds_client.create_index(index, settings=parameters[-1], mappings=parameters[0])
    doc = [
        {
            "_index": index,
            "_id": i,
            "_source": {
                "article_id": i,
                "text": f"Amazing text {i}",
                "paragraph_id": i,
            },
        }
        for i in range(10)
    ]

    await ds_client.bulk(doc)
    await ds_client.client.indices.refresh()
    n_doc = await ds_client.count_documents(index=index)
    assert n_doc == 10
    retrieved_doc = await ds_client.get_document(index=index, doc_id=3)
    assert retrieved_doc["text"] == "Amazing text 3"


@pytest.mark.asyncio
@pytest.mark.parametrize("query", [{"match_all": {}}, {"match": {"text": "retrieve"}}])
async def test_asearch(get_testing_async_ds_client, query):
    """Test the retrieval of a document."""
    ds_client, parameters = get_testing_async_ds_client

    aggs = {"unique_ids": {"terms": {"field": "paragraph_id"}}}
    index_doc = "test_paragraphs"

    await ds_client.create_index(
        index_doc, settings=parameters[-1], mappings=parameters[0]
    )

    doc_1 = {
        "_index": index_doc,
        "_id": 1,
        "_source": {
            "text": "test of an amazing function",
            "title": "test_article",
            "paragraph_id": "1",
            "article_id": "article_id",
            "journal": "8765-4321",
        },
    }

    doc_2 = {
        "_index": index_doc,
        "_id": 2,
        "_source": {
            "text": "The bird sings very loudly",
            "title": "test_article",
            "paragraph_id": "1",
            "article_id": "article_id",
            "journal": None,
        },
    }

    doc_3 = {
        "_index": index_doc,
        "_id": 3,
        "_source": {
            "text": "This document is a bad test, I don't want to retrieve it",
            "title": "test_article",
            "paragraph_id": "1",
            "article_id": "article_id",
            "journal": "1234-5678",
        },
    }
    doc_bulk = [doc_1, doc_2, doc_3]

    await ds_client.bulk(doc_bulk)

    await ds_client.client.indices.refresh()

    results = await ds_client.search(index_doc, query, aggs=aggs)
    res_hits = results["hits"]["hits"]
    assert res_hits[0]["_source"]["article_id"] == "article_id"
    assert res_hits[0]["_source"]["title"] == "test_article"
    if "match_all" in query.keys():
        assert len(res_hits) == 3
        assert res_hits[0]["_source"]["text"] == "test of an amazing function"
        assert res_hits[0]["_score"] == 1.0

        # test query + aggs.
        results = await ds_client.search(index_doc, query, aggs=aggs)
        assert results["aggregations"]["unique_ids"]["buckets"][0]["doc_count"] == 3
    else:
        assert len(res_hits) == 1
        assert (
            res_hits[0]["_source"]["text"]
            == "This document is a bad test, I don't want to retrieve it"
        )
        assert res_hits[0]["_score"] == pytest.approx(0.7782316)
        # test query + aggs.
        results = await ds_client.search(index_doc, query, aggs=aggs)
        assert results["aggregations"]["unique_ids"]["buckets"][0]["doc_count"] == 1

    # test aggs alone.
    results = await ds_client.search(index_doc, aggs=aggs)
    assert results["aggregations"]["unique_ids"]["buckets"][0]["doc_count"] == 3

    # test without anything.
    results = await ds_client.search(index_doc)
    assert "aggregations" not in results.keys()
    assert len(results["hits"]["hits"]) == 3


@pytest.mark.asyncio
@pytest.mark.parametrize("filter_db", [None, {"match": {"text": "retrieve"}}])
async def test_abm25_search(get_testing_async_ds_client, filter_db):
    """Test the retrieval of a document."""
    ds_client, parameters = get_testing_async_ds_client

    query = "test"
    index_doc = "test_paragraphs"

    await ds_client.create_index(
        index_doc, settings=parameters[-1], mappings=parameters[0]
    )

    doc_1 = {
        "_index": index_doc,
        "_id": 1,
        "_source": {
            "text": "test of an amazing function",
            "title": "test_article",
            "paragraph_id": "1",
            "article_id": "article_id",
            "journal": "8765-4321",
        },
    }

    doc_2 = {
        "_index": index_doc,
        "_id": 2,
        "_source": {
            "text": "The bird sings very loudly",
            "title": "test_article",
            "paragraph_id": "1",
            "article_id": "article_id",
            "journal": None,
        },
    }

    doc_3 = {
        "_index": index_doc,
        "_id": 3,
        "_source": {
            "text": "This document is a bad test, I don't want to retrieve it",
            "title": "test_article",
            "paragraph_id": "1",
            "article_id": "article_id",
            "journal": "1234-5678",
        },
    }
    doc_bulk = [doc_1, doc_2, doc_3]

    await ds_client.bulk(doc_bulk)

    await ds_client.client.indices.refresh()

    res = await ds_client.bm25_search(index_doc, query, filter_db, 2)
    assert res[0]["article_id"] == "article_id"
    assert res[0]["title"] == "test_article"
    if filter_db is None:
        assert len(res) == 2
        assert res[0]["text"] == "test of an amazing function"
        assert res[0]["score"] == pytest.approx(0.5403367)
    else:
        assert len(res) == 1
        assert (
            res[0]["text"] == "This document is a bad test, I don't want to retrieve it"
        )
        assert res[0]["score"] == pytest.approx(0.37292093)


@pytest.mark.parametrize(
    "query, expected_query",
    [
        ({"regex": {"query": "test:test"}}, {"regex": {"query": "test:test"}}),
        (
            {"query_string": {"query": "test:test"}},
            {"query_string": {"query": r"test\:test"}},
        ),
        (
            {"query_string": {"query": "test^test"}},
            {"query_string": {"query": r"test\^test"}},
        ),
        (
            {"query_string": {"query": "test>test"}},
            {"query_string": {"query": r"testtest"}},
        ),
        (
            {"query_string": {"query": "test<test"}},
            {"query_string": {"query": r"testtest"}},
        ),
        (None, None),
    ],
)
def test_postprocess_query(query, expected_query):
    postprocessed_query = postprocess_query(query)
    assert postprocessed_query == expected_query



"""Consumer for parse and upload."""

import argparse
import asyncio
import logging
import os
import pathlib
import time
from multiprocessing import Process, cpu_count
from typing import Any

from aiobotocore.session import get_session
from botocore import UNSIGNED
from botocore.config import Config
from elasticsearch import ApiError
from elasticsearch.helpers import BulkIndexError as ESBulkIndexError
from httpx import AsyncClient
from opensearchpy.exceptions import TransportError
from opensearchpy.helpers import BulkIndexError as OSBulkIndexError

from scholarag.ds_utils import ds_upload, setup_parsing_ds

logger = logging.getLogger(__name__)


def get_parser() -> argparse.ArgumentParser:
    """Get parser for command line arguments."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "db_url",
        type=str,
        help="URL of the database instance (together with the port).",
    )
    parser.add_argument(
        "parser_url",
        type=str,
        help="URL of the parser (needs to contain the host:port).",
    )
    parser.add_argument(
        "queue_url",
        type=str,
        help="Url of the queue where to upload the articles.",
    )
    parser.add_argument(
        "--batch-size",
        "-b",
        type=int,
        default=500,
        help="Maximum number of articles to process before sending them to the db.",
    )
    parser.add_argument(
        "--max-paragraphs-length",
        "-u",
        type=int,
        help="Maximum length a paragraph is allowed to have to be uploaded to the DB.",
        default=None,
    )
    parser.add_argument(
        "--min-paragraphs-length",
        "-l",
        type=int,
        help="Minimum length a paragraph is allowed to have to be uploaded to the DB.",
        default=None,
    )
    parser.add_argument(
        "--max-concurrent-requests",
        type=int,
        default=10,
        help=(
            "Maximum number of articles sent to the etl server PER BATCH OF"
            " 'articles-per-bulk' articles."
        ),
    )
    parser.add_argument(
        "--db-type",
        type=str,
        choices=["elasticsearch", "opensearch"],
        help="Type of database.",
        default="opensearch",
        required=False,
    )
    parser.add_argument(
        "--user",
        type=str,
        default=None,
        help=(
            "User of the database instance. The password is read from the environment"
            " under SCHOLARAG__DB__PASSWORD."
        ),
        required=False,
    )
    parser.add_argument(
        "--use-ssl",
        action="store_true",
        help="Whether to verify ssl certificates or not.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Control verbosity",
    )
    return parser


async def run(
    worker_n: int,
    db_url: str,
    parser_url: str,
    queue_url: str,
    db_type: str,
    max_concurrent_requests: int = 10,
    min_paragraphs_length: int | None = None,
    max_paragraphs_length: int | None = None,
    batch_size: int = 500,
    user: str | None = None,
    use_ssl: bool = False,
) -> None:
    """Run the consumer script.

    Parameters
    ----------
    worker_n
        Number of workers.
    db_url
        URL of the db to which the documents should be sent
    parser_url
        Base url of the parser (without the endpoint).
    queue_url
        URL of the queue that holds the articles.
    db_type
        Either elasticsearch of opensearch.
    max_concurrent_requests
        Maximum number of concurrent requests sent to the etl service.
    min_paragraphs_length
        Minimum length a paragraph is allowed to have to be uploaded to the DB.
    max_paragraphs_length
        Maximum length a paragraph is allowed to have to be uploaded to the DB.
    batch_size
        Number of articles to retrieve from the queue at every iteration.
    user
        Username to authenticate in the DB.
    use_ssl
        Whether to verify ssl certificate or not.
    """
    try:
        password = os.getenv("SCHOLARAG__DB__PASSWORD")
        if not password:
            logger.warning(
                f"[WORKER {worker_n}] SCHOLARAG__DB__PASSWORD variable not set."
                " defaulting to None."
            )
        # Setup parser and db.
        ds_client, parser_service = await setup_parsing_ds(
            db_url,
            db_type,
            user,
            password,
            max_concurrent_requests,
            use_ssl,
        )
        session = get_session()
        httpx_client = AsyncClient(timeout=60)
        tested_indices: set[str] = set()

        async with (
            session.create_client(
                "s3",
                region_name="us-east-1",
            ) as signed_s3_client,
            session.create_client("sqs", region_name="us-east-1") as q_client,
            session.create_client(
                "s3",
                region_name="us-east-1",
                config=Config(signature_version=UNSIGNED),
            ) as unsigned_s3_client,
            httpx_client as httpx_client,
        ):
            # Iterate while the queue is not empty.
            while True:
                start_time_batch = time.time()
                batch: list[dict[str, Any]] = []

                # Retrieve a batch from the queue.
                logger.info(f"[WORKER {worker_n}] Polling the queue...")
                attempt = 0
                while (d_batch := batch_size - len(batch)) > 0:
                    messages = await q_client.receive_message(
                        QueueUrl=queue_url,
                        WaitTimeSeconds=2,
                        MaxNumberOfMessages=10 if d_batch > 10 else d_batch,
                        MessageAttributeNames=[
                            "Key",
                            "Bucket_Name",
                            "Parser_Endpoint",
                            "Sign",
                            "Index",
                        ],
                    )
                    if "Messages" not in messages:
                        # If there is nothing in the queue left, retry.
                        attempt += 1
                        # If there was not enough in the queue to fill a batch, run with partial batch if it is not empty.
                        if attempt >= 10 and batch:
                            logger.info(
                                f"[WORKER {worker_n}] Managed to retrieve a partial"
                                f" batch of len {len(batch)}."
                            )
                            break
                        continue
                    batch.extend(messages["Messages"])

                indices = [
                    message["MessageAttributes"]["Index"]["StringValue"]
                    for message in batch
                ]
                indices_to_test = set(indices) - tested_indices
                for index_to_test in indices_to_test:
                    if index_to_test not in await ds_client.get_available_indexes():
                        raise RuntimeError(
                            f"Index {index_to_test} not in document store. Create it"
                            " before pushing articles to it"
                        )
                    else:
                        tested_indices.add(index_to_test)

                files = []
                # Download relevant articles.
                for message in batch:
                    client = (
                        signed_s3_client
                        if bool(message["MessageAttributes"]["Sign"]["StringValue"])
                        else unsigned_s3_client
                    )
                    response = await client.get_object(
                        Bucket=message["MessageAttributes"]["Bucket_Name"][
                            "StringValue"
                        ],
                        Key=message["MessageAttributes"]["Key"]["StringValue"],
                    )
                    body = await response["Body"].read()
                    files.append(body)

                # Parse the articles.
                logger.info(
                    f"[WORKER {worker_n}] Request server to parse"
                    f" {len(batch)} documents."
                )
                filenames = [
                    pathlib.Path(message["MessageAttributes"]["Key"]["StringValue"])
                    for message in batch
                ]

                # Messages do not necessarily use the same parser, so we send them one by one with their own url.
                results = []
                for file, message in zip(files, batch):
                    url = (
                        parser_url
                        + f'/{message["MessageAttributes"]["Parser_Endpoint"]["StringValue"]}'
                    )
                    result = await parser_service.arun(
                        files=[file],
                        url=url,
                        multipart_params={} if "grobid" in url else None,
                        httpx_client=httpx_client,
                    )
                    results.extend(result)
                logger.info(
                    f"[WORKER {worker_n}] Collecting data to upload to the document"
                    " store."
                )
                # Try to upload them to the db. If successful, remove from queue.
                try:
                    await ds_upload(
                        filenames=filenames,
                        results=results,
                        ds_client=ds_client,
                        min_paragraphs_length=min_paragraphs_length,
                        max_paragraphs_length=max_paragraphs_length,
                        indices=indices,
                    )
                    to_delete_from_q = [
                        {
                            "Id": message["MessageId"],
                            "ReceiptHandle": message["ReceiptHandle"],
                        }
                        for (message, result) in zip(batch, results)
                        if result is not None
                    ]
                except (ApiError, ESBulkIndexError, TransportError, OSBulkIndexError):
                    to_delete_from_q = []
                    logger.info(f"[WORKER {worker_n}] Failed parsing the batch.")

                logger.info(
                    f"[WORKER {worker_n}] Batch took {time.time() - start_time_batch}"
                )
                # Remove successful articles from queue.
                for i in range(0, len(to_delete_from_q), 10):
                    response = await q_client.delete_message_batch(
                        QueueUrl=queue_url,
                        Entries=to_delete_from_q[
                            i : min(i + 10, len(to_delete_from_q))
                        ],
                    )
                    if "Failed" in response:
                        logger.info(f"[WORKER {worker_n}] {response['Failed']}")
                    else:
                        logger.info(
                            f"[WORKER {worker_n}] Successfully deleted"
                            f" {len(to_delete_from_q[i:min(i+10, len(to_delete_from_q))])} entries"
                            " from the queue."
                        )
    except BaseException as e:
        logger.info(f"[WORKER {worker_n}] Closed queue polling. {e}")
        await ds_client.close()
        raise e


def sub_process(
    worker_n: int,
    db_url: str,
    parser_url: str,
    queue_url: str,
    db_type: str,
    max_concurrent_requests: int = 10,
    batch_size: int = 500,
    min_paragraphs_length: int | None = None,
    max_paragraphs_length: int | None = None,
    user: str | None = None,
    use_ssl: bool = False,
    verbose: bool = False,
) -> None:
    """Run the main logic."""
    # Additional parameters
    logging_level = logging.INFO if verbose else logging.WARNING

    # setup logging
    logging.basicConfig(
        format="[%(levelname)s]  %(asctime)s %(name)s  %(message)s", level=logging_level
    )
    raise SystemExit(
        asyncio.run(
            run(
                worker_n=worker_n,
                db_url=db_url,
                parser_url=parser_url,
                queue_url=queue_url,
                db_type=db_type,
                max_concurrent_requests=max_concurrent_requests,
                batch_size=batch_size,
                min_paragraphs_length=min_paragraphs_length,
                max_paragraphs_length=max_paragraphs_length,
                user=user,
                use_ssl=use_ssl,
            )
        )
    )


def main() -> None:
    """Run multiple sub_processes."""
    # Parse command line
    parser = get_parser()
    args = parser.parse_args()
    for i in range(cpu_count()):
        p = Process(target=sub_process, kwargs={"worker_n": i + 1, **args.__dict__})
        p.start()


if __name__ == "__main__":
    main()



"""Document Store utilities."""

from __future__ import annotations

import hashlib
import logging
import math
import pathlib
from typing import Any

from elasticsearch import ApiError
from elasticsearch.helpers import BulkIndexError as ESBulkIndexError
from opensearchpy.exceptions import TransportError
from opensearchpy.helpers import BulkIndexError as OSBulkIndexError

from scholarag.document_stores import (
    AsyncBaseSearch,
    AsyncElasticSearch,
    AsyncOpenSearch,
)
from scholarag.services import ParsingService
from scholarag.utils import find_files

logger = logging.getLogger("ds_utils")
logger_os = logging.getLogger("opensearch")
logger_es = logging.getLogger("elasticsearch")


class NoParsingFilter(logging.Filter):
    """Class to filter logs coming from ds_client.exists()."""

    def filter(self, record: logging.LogRecord) -> bool:  # noqa: A003
        """Filter logs coming from ds_client.exists()."""
        return not record.getMessage().startswith("GET")


logger_os.addFilter(NoParsingFilter())
logger_es.addFilter(NoParsingFilter())


async def check_docs_exists_in_db(
    client: AsyncBaseSearch, index: str, pmc_ids: list[str]
) -> list[str]:
    """Return list of index of documents .

    Parameters
    ----------
    client
        Async Opensearch client.
    index
        OS index where documents are stored.
    pmc_ids
        PMC IDs of the documents.

    Returns
    -------
        True if the document exists within an index.
    """
    query = {"query": {"terms": {"pmc_id": pmc_ids}}}
    aggs = {"matched_ids": {"terms": {"field": "pmc_id", "size": len(pmc_ids)}}}
    docs = await client.search(index=index, query=query, aggs=aggs, size=0)
    existing_ids = [
        doc["key"] for doc in docs["aggregations"]["matched_ids"]["buckets"]
    ]
    return existing_ids


async def setup_parsing_ds(
    db_url: str,
    db_type: str = "opensearch",
    user: str | None = None,
    password: str | None = None,
    max_concurrent_requests: int = 10,
    use_ssl: bool = False,
) -> tuple[AsyncBaseSearch, ParsingService]:
    """Create the ds client and parsing service.

    Parameters
    ----------
    db_url
        Url of the DB
    db_type
        Opensearch or Elasticsearch
    user
        Username to connect to the DB
    password
        Password to connect to the DB
    max_concurrent_requests
        Max number of requests to send simultaneously to the parser
    use_ssl
        Whether certificates should be verified or not

    Returns
    -------
    ds_client, parsing_service
    """
    # Setup document store.
    host, _, port = db_url.rpartition(":")
    ds_client = (
        AsyncOpenSearch(  # type:ignore
            host=host,
            port=int(port),
            user=user,
            password=password,
            use_ssl_and_verify_certs=use_ssl,
        )
        if db_type == "opensearch"
        else AsyncElasticSearch(  # type: ignore
            host=host,
            port=int(port),
            user=user,
            password=password,
            use_ssl_and_verify_certs=use_ssl,
        )
    )

    # Setup parsing service
    parsing_service = ParsingService(
        ignore_errors=True,
        max_concurrent_requests=max_concurrent_requests,
    )
    return ds_client, parsing_service


def get_files(
    path: pathlib.Path,
    recursive: bool,
    articles_per_bulk: int,
    match_filename: str | None = None,
) -> list[list[pathlib.Path]]:
    """Fetch local files to parser and upload."""
    # Check the path and iterate to have a list of papers to parse
    if not path.exists():
        raise ValueError(f"The path {path} does not exist.")

    if path.is_dir():
        inputs = find_files(path, recursive, match_filename)
    else:
        inputs = [
            path,
        ]

    logger.info(
        f"There are {len(inputs)} files to parse, split in"
        f" {math.ceil(len(inputs) / articles_per_bulk)} batches."
    )
    batches = [
        inputs[i : i + articles_per_bulk]
        for i in range(0, len(inputs), articles_per_bulk)
    ]
    return batches


async def ds_upload(
    filenames: list[pathlib.Path],
    results: list[dict[str, Any] | None],
    indices: list[str],
    ds_client: AsyncBaseSearch,
    min_paragraphs_length: int | None = None,
    max_paragraphs_length: int | None = None,
) -> list[pathlib.Path]:
    """Upload results to document store.

    Parameters
    ----------
    filenames
        List of files from which the content is uploaded
    results
        Output of the parser
    indices
        List of index to which each document belong to.
    ds_client
        Document store client
    min_paragraphs_length
        Minimum length a paragraph is allowed to have to be uploaded to the DB.
    max_paragraphs_length
        Maximum length a paragraph is allowed to have to be uploaded to the DB.

    Returns
    -------
    List of files that failed to be pushed.
    """
    upload_bulk = []
    files_failing: list[pathlib.Path] = []
    doc_ids: set[str] = set()
    for k, (res, index) in enumerate(zip(results, indices)):
        if res is None:
            logging.error(f"File {filenames[k]} failed to be parsed.")
            files_failing.append(filenames[k])
            continue
        try:
            for i, abstract in enumerate(res["abstract"]):
                doc_id = hashlib.md5(
                    (res["uid"] + abstract).encode("utf-8"),
                    usedforsecurity=False,
                ).hexdigest()
                # Check length (not expensive) before doing a db call to check existence.
                if (
                    min_paragraphs_length and len(abstract) <= min_paragraphs_length
                ) or (max_paragraphs_length and len(abstract) >= max_paragraphs_length):
                    continue
                # if length is ok, check if it already exists somewhere.
                if doc_id in doc_ids or await ds_client.exists(index, doc_id):
                    continue
                else:
                    doc_ids.add(doc_id)

                par = {
                    "_index": index,
                    "_id": doc_id,
                    "_source": {
                        "article_id": res["uid"],
                        "section": "Abstract",
                        "text": abstract,
                        "paragraph_id": i,
                        "authors": res["authors"],
                        "title": res["title"],
                        "pubmed_id": res["pubmed_id"],
                        "pmc_id": res["pmc_id"],
                        "arxiv_id": res["arxiv_id"],
                        "doi": res["doi"],
                        "date": res["date"],
                        "journal": res["journal"],
                        "article_type": res["article_type"],
                    },
                }
                if len(par) > 0:
                    upload_bulk.append(par)

            for ppos, (section, text) in enumerate(res["section_paragraphs"]):
                doc_id = hashlib.md5(
                    (res["uid"] + text).encode("utf-8"), usedforsecurity=False
                ).hexdigest()
                # Check length (not expensive) before doing a db call to check existence.
                if (min_paragraphs_length and len(text) <= min_paragraphs_length) or (
                    max_paragraphs_length and len(text) >= max_paragraphs_length
                ):
                    continue
                # if length is ok, check if it already exists.
                if doc_id in doc_ids or await ds_client.exists(index, doc_id):
                    continue
                else:
                    doc_ids.add(doc_id)

                par = {
                    "_index": index,
                    "_id": doc_id,
                    "_source": {
                        "article_id": res["uid"],
                        "section": section,
                        "text": text,
                        "paragraph_id": ppos + len(res["abstract"]),
                        "authors": res["authors"],
                        "title": res["title"],
                        "pubmed_id": res["pubmed_id"],
                        "pmc_id": res["pmc_id"],
                        "arxiv_id": res["arxiv_id"],
                        "doi": res["doi"],
                        "date": res["date"],
                        "journal": res["journal"],
                        "article_type": res["article_type"],
                    },
                }
                if len(par) > 0:
                    upload_bulk.append(par)
        except TypeError as e:
            logger.error(f"Article {filenames[k]} could not be parsed properly. {e}")
            files_failing.append(filenames[k])
            continue

    logger.info("Upload data to database.")
    try:
        await ds_client.bulk(upload_bulk)
    except (ApiError, ESBulkIndexError, TransportError, OSBulkIndexError) as e:
        logger.info(f"Results could not be uploaded properly. {e}")
        logger.info(f"List of files that failed: {filenames}")
        raise e

    return files_failing



"""Parse articles and upload them on the document store."""

from __future__ import annotations

import argparse
import asyncio
import json
import logging
import os
import pathlib
import time
from datetime import datetime, timedelta
from typing import Any

from aiobotocore.session import get_session
from botocore import UNSIGNED
from botocore.config import Config
from elasticsearch import ApiError
from elasticsearch.helpers import BulkIndexError as ESBulkIndexError
from httpx import AsyncClient
from opensearchpy.exceptions import TransportError
from opensearchpy.helpers import BulkIndexError as OSBulkIndexError

from scholarag.ds_utils import check_docs_exists_in_db, ds_upload, setup_parsing_ds

logger = logging.getLogger(__name__)


def get_parser() -> argparse.ArgumentParser:
    """Get parser for command line arguments."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "db_url",
        type=str,
        help="URL of the database instance (together with the port).",
    )
    parser.add_argument(
        "parser_url",
        type=str,
        help="URL of the parser (needs to contain the host:port and the parser type).",
    )
    parser.add_argument(
        "--start-date",
        type=str,
        default=None,
        help=(
            "Date of the oldest document to download. Defaults to yesterday. Format:"
            " dd-mm-yyy."
        ),
    )
    parser.add_argument(
        "--max-paragraphs-length",
        "-u",
        type=int,
        help="Maximum length a paragraph is allowed to have to be uploaded to the DB.",
        default=None,
    )
    parser.add_argument(
        "--min-paragraphs-length",
        "-l",
        type=int,
        help="Minimum length a paragraph is allowed to have to be uploaded to the DB.",
        default=None,
    )
    parser.add_argument(
        "--batch-size",
        type=int,
        default=500,
        help="Maximum number of articles to process before sending them to the db.",
    )
    parser.add_argument(
        "--multipart-params",
        default=None,
        type=json.loads,
        help=(
            "Dictionary for the multipart parameters for parsers that require some."
            " Must be passed as a json formated dict."
        ),
    )
    parser.add_argument(
        "--max-concurrent-requests",
        type=int,
        default=10,
        help=(
            "Maximum number of articles sent to the etl server PER BATCH OF"
            " 'articles-per-bulk' articles."
        ),
    )
    parser.add_argument(
        "--db-type",
        type=str,
        choices=["elasticsearch", "opensearch"],
        help="Type of database.",
        default="opensearch",
        required=False,
    )
    parser.add_argument(
        "--user",
        type=str,
        default=None,
        help=(
            "User of the database instance. The password is read from the environment"
            " under SCHOLARAG__DB__PASSWORD."
        ),
        required=False,
    )
    parser.add_argument(
        "--index",
        type=str,
        default="pmc_paragraphs",
        help="Desired name of the index holding paragraphs.",
    )
    parser.add_argument(
        "--files-failing-path",
        type=pathlib.Path,
        default=None,
        help=(
            "Path where to dump the files that failed. Expects a file. Not dumping if"
            " None."
        ),
    )
    parser.add_argument(
        "--use-ssl",
        action="store_true",
        help="Whether to verify ssl certificates or not.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Control verbosity",
    )
    return parser


async def run(
    db_url: str,
    parsing_url: str,
    start_date: datetime,
    db_type: str,
    multipart_params: dict[str, Any] | None = None,
    max_concurrent_requests: int = 10,
    index: str = "pmc_paragraphs",
    batch_size: int = 500,
    min_paragraphs_length: int | None = None,
    max_paragraphs_length: int | None = None,
    user: str | None = None,
    use_ssl: bool = False,
    files_failing_path: pathlib.Path | None = None,
) -> int:
    """Run the article parsing and upload results on DS."""
    password = os.getenv("SCHOLARAG__DB__PASSWORD")
    if not password:
        logger.warning("SCHOLARAG__DB__PASSWORD variable not set. defaulting to None.")
    ds_client, parsing_service = await setup_parsing_ds(
        db_url,
        db_type,
        user,
        password,
        max_concurrent_requests,
        use_ssl,
    )
    # Setup s3 bucket
    logger.info("Starting S3 client.")
    session = get_session()
    start_time = time.time()
    async with (
        session.create_client(
            "s3", region_name="us-east-1", config=Config(signature_version=UNSIGNED)
        ) as client,
        AsyncClient() as httpx_client,
    ):
        for prefix in [
            "oa_comm/xml/all/",
            "oa_noncomm/xml/all/",
            "author_manuscript/xml/all/",
        ]:
            s3_paginator = client.get_paginator("list_objects_v2")
            s3_iterator = s3_paginator.paginate(Bucket="pmc-oa-opendata", Prefix=prefix)
            logger.info("Filtering interesting articles.")
            filtered_iterator = s3_iterator.search(
                f"""Contents[?to_string(LastModified)>='\"{start_date.strftime('%Y-%m-%d %H:%M:%S%')}+00:00\"'
                    && contains(Key, '.xml')]"""
            )
            finished = False
            while not finished:
                start_time_batch = time.time()

                s3_bucket_dict = {}
                i = 0
                # Creating batches.
                async for obj in filtered_iterator:
                    i += 1
                    s3_bucket_key = pathlib.Path(obj["Key"])
                    s3_bucket_dict[s3_bucket_key.stem] = s3_bucket_key
                    if i == batch_size:
                        break

                if i != batch_size:
                    finished = True

                all_pmc_ids = list(s3_bucket_dict.keys())
                if not all_pmc_ids:
                    continue
                existing_files = await check_docs_exists_in_db(
                    ds_client, index, all_pmc_ids
                )

                s3_bucket_keys = [
                    v for k, v in s3_bucket_dict.items() if k not in existing_files
                ]

                logger.info("Get objects.")
                files = []
                for s3_bucket_key in s3_bucket_keys:
                    response = await client.get_object(
                        Bucket="pmc-oa-opendata", Key=str(s3_bucket_key)
                    )
                    body = await response["Body"].read()
                    files.append(body)

                logger.info(f"Request server to parse {len(s3_bucket_keys)} documents.")
                results = await parsing_service.arun(
                    url=parsing_url,
                    files=files,
                    multipart_params=multipart_params,
                    httpx_client=httpx_client,
                )
                logger.info("Collecting data to upload to the document store.")
                indices = [index for _ in range(len(s3_bucket_keys))]
                try:
                    files_failing = await ds_upload(
                        filenames=s3_bucket_keys,
                        results=results,
                        ds_client=ds_client,
                        min_paragraphs_length=min_paragraphs_length,
                        max_paragraphs_length=max_paragraphs_length,
                        indices=indices,
                    )
                except (ApiError, ESBulkIndexError, TransportError, OSBulkIndexError):
                    files_failing = s3_bucket_keys

                if files_failing and files_failing_path is not None:
                    logger.info(
                        f"Dumping failing files to {files_failing_path.resolve()}"
                    )
                    with open(files_failing_path, "a") as f:
                        for file in files_failing:
                            f.write(json.dumps(file))
                            f.write("\n")
                    logger.info(
                        f"From start took: {time.time() - start_time}. Batch took:"
                        f" {time.time() - start_time_batch}"
                    )

    logger.info("Done.")

    return 0


def main() -> None:
    """Run the main logic."""
    # Parse command line
    parser = get_parser()
    args = parser.parse_args()

    # Additional parameters
    logging_level = logging.INFO if args.verbose else logging.WARNING

    # setup logging
    logging.basicConfig(
        format="[%(levelname)s]  %(asctime)s %(name)s  %(message)s", level=logging_level
    )
    if args.start_date is not None:
        start_date = datetime.strptime(args.start_date, "%d-%m-%Y")
        if start_date > datetime.today():
            raise ValueError(
                "Date is in the future! Hope you found the secret for time travelling."
            )
    else:
        start_date = datetime.today() - timedelta(days=1)
    raise SystemExit(
        asyncio.run(
            run(
                db_url=args.db_url,
                parsing_url=args.parser_url,
                start_date=start_date,
                db_type=args.db_type,
                user=args.user,
                index=args.index,
                min_paragraphs_length=args.min_paragraphs_length,
                max_paragraphs_length=args.max_paragraphs_length,
                multipart_params=args.multipart_params,
                max_concurrent_requests=args.max_concurrent_requests,
                use_ssl=args.use_ssl,
                batch_size=args.batch_size,
                files_failing_path=args.files_failing_path,
            )
        )
    )


if __name__ == "__main__":
    main()



"""Services."""

from scholarag.services.cohere_reranker import CohereRerankingService
from scholarag.services.etl import ParsingService
from scholarag.services.retrieval import RetrievalService

__all__ = [
    "CohereRerankingService",
    "ParsingService",
    "RetrievalService",
]



"""Utilities to stream openai response."""

import json
from collections import deque
from typing import Any, AsyncIterable

from httpx import AsyncClient
from openai import AsyncOpenAI, BadRequestError

from scholarag.app.config import Settings
from scholarag.app.dependencies import ErrorCode
from scholarag.document_stores import AsyncBaseSearch
from scholarag.generative_question_answering import (
    ERROR_SEPARATOR,
    SOURCES_SEPARATOR,
    GenerativeQAWithSources,
)
from scholarag.retrieve_metadata import MetaDataRetriever


async def stream_response(
    qas: GenerativeQAWithSources,
    query: str,
    contexts_text: list[str],
    ds_client: AsyncBaseSearch,
    index_journals: str | None,
    index_paragraphs: str,
    metadata_retriever: MetaDataRetriever,
    settings: Settings,
    contexts: list[dict[str, Any]],
    indices: tuple[int, ...],
    scores: tuple[float, ...] | None,
) -> AsyncIterable[str]:
    """Generate the stream for the `streamed_generative` endpoint.

    Parameters
    ----------
    qas
        Generative QA class to run the model's inference.
    query
        Question of the user.
    contexts_text
        list of texts to run the qa against.
    ds_client
        Document store client.
    index_journals
        DB index containing the impact factors.
    index_paragraphs
        DB index containing the paragraphs.
    metadata_retriever
        Class used to retreave metadata on the fly.
    settings
        Settings of the applications.
    contexts
        Contexts as extracted from the DB (with metadata).
    indices
        Order of the contexts after (potential) reranking.
    scores
        Reranking scores.

    Yields
    ------
    Generated tokens on the fly, + entire response in the end or error messages.
    """
    # Begin a task that runs in the background.
    interrupted = False

    # Needed because of fastAPI closing connections.
    httpx_client = AsyncClient(
        timeout=settings.metadata.timeout, verify=False
    )  # Using dependencies doesn't work here.
    password = settings.db.password.get_secret_value() if settings.db.password else None
    api_key = (
        settings.generative.openai.token.get_secret_value()
        if settings.generative.openai.token
        else None
    )
    ds_client = ds_client.__class__(
        host=settings.db.host,
        port=settings.db.port,
        user=settings.db.user,
        password=password,
    )
    openai_client = AsyncOpenAI(api_key=api_key)
    qas.client = openai_client

    token_queue: deque[str] = deque(
        maxlen=4
    )  # Needs to be EXACTLY the number of tokens of the separator. See https://platform.openai.com/tokenizer
    # Some black magic.
    raw_string = ""
    try:
        async for chunk in qas.astream(
            query=query,
            contexts=contexts_text,
            prompt_template=settings.generative.prompt_template.get_secret_value(),
        ):
            raw_string += chunk
            if len(token_queue) == token_queue.maxlen:
                queued_text = "".join(list(token_queue))
                if (
                    f"{SOURCES_SEPARATOR}:" in queued_text
                    or ERROR_SEPARATOR in queued_text
                ):  # Might change if we change the separator. # This condition is subject to change based on the separator we use.
                    continue
                yield token_queue.popleft()
            token_queue.append(chunk)
    except RuntimeError as e:
        # Since finish reason is raised, it has to be recovered as such.
        finish_reason = e.args[0]

    # Errors cannot be raised due to the streaming nature of the endpoint. They're yielded as a dict instead,
    # leaving the charge to the user to catch them. The stream is simply interrupted in case of error.
    except BadRequestError as err:
        interrupted = True
        # Adding a separator before raising the error
        yield "<bbs_json_error>"
        yield json.dumps(
            {
                "Error": {
                    "status_code": 413,
                    "code": ErrorCode.INPUT_EXCEEDED_MAX_TOKENS.value,
                    "detail": err.body["message"],  # type: ignore
                }
            }
        )
    if not interrupted:
        # Post process the output to get citations.
        answer = qas._process_raw_output(raw_output=raw_string)
        # Ensure that the model finished yielding.
        if finish_reason == "length":
            # Adding a separator before raising the error.
            yield "<bbs_json_error>"
            yield json.dumps(
                {
                    "Error": {
                        "status_code": 500,
                        "code": ErrorCode.INCOMPLETE_ANSWER.value,
                        "detail": (
                            "The LLM did not have enough completion tokens available to"
                            " finish its answer. Please decrease the reranker_k or"
                            " retriever_k value by 1 or 2 depending of whether you are"
                            " using the reranker or not."
                        ),
                        "raw_answer": answer["raw_answer"],
                    }
                }
            )
        elif answer["paragraphs"] is None or len(answer["paragraphs"]) == 0:  # type: ignore
            # Adding a separator before raising the error.
            yield "<bbs_json_error>"
            yield json.dumps(
                {
                    "Error": {
                        "status_code": 404,
                        "code": ErrorCode.NO_ANSWER_FOUND.value,
                        "detail": (
                            "The LLM did not provide any source to answer the question."
                        ),
                        "raw_answer": answer["raw_answer"],
                    }
                }
            )
        else:
            # Adding a separator between the streamed answer and the processed response.
            yield "<bbs_json_data>"
            complete_answer = await retrieve_metadata(
                answer=answer,
                ds_client=ds_client,
                index_journals=index_journals,
                index_paragraphs=index_paragraphs,
                metadata_retriever=metadata_retriever,
                httpx_client=httpx_client,
                contexts=contexts,
                indices=indices,
                scores=scores,
            )
            yield json.dumps(complete_answer)


async def retrieve_metadata(
    answer: dict[str, Any],
    ds_client: AsyncBaseSearch,
    index_journals: str | None,
    index_paragraphs: str,
    metadata_retriever: MetaDataRetriever,
    httpx_client: AsyncClient,
    contexts: list[dict[str, Any]],
    indices: tuple[int, ...],
    scores: tuple[float, ...] | None,
) -> dict[str, Any]:
    """Retrieve the metadata and display them nicely.

    Parameters
    ----------
    answer
        Answer generated by the model.
    ds_client
        Document store client.
    index_journals
        DB index containing the impact factors.
    index_paragraphs
        DB index containing the paragraphs.
    metadata_retriever
        Class used to retreave metadata on the fly.
    httpx_client
        HTTPX Client.
    contexts
        Contexts as extracted from the DB (with metadata).
    indices
        Order of the contexts after (potential) reranking.
    scores
        Reranking scores.

    Returns
    -------
    Nicely formatted answer with metadata.
    """
    fetched_metadata = await metadata_retriever.retrieve_metadata(
        contexts, ds_client, index_journals, index_paragraphs, httpx_client
    )

    citedby_counts = fetched_metadata.get("get_citation_count", {})
    journal_names = fetched_metadata.get("get_journal_name", {})
    impact_factors = fetched_metadata["get_impact_factors"]
    abstracts = fetched_metadata["recreate_abstract"]

    context_ids: list[int] = answer["paragraphs"]
    metadata = []
    for context_id in context_ids:
        context = contexts[context_id]

        metadata.append(
            {
                "impact_factor": impact_factors[context.get("journal")],
                "journal_name": journal_names.get(context.get("journal")),
                "journal_issn": context.get("journal"),
                "date": context.get("date"),
                "cited_by": citedby_counts.get(context.get("doi")),
                "article_id": context["article_id"],
                "ds_document_id": context["document_id"],
                "context_id": indices[context_id],
                "reranking_score": scores[context_id] if scores is not None else None,
                "article_doi": context["doi"],
                "article_title": context["title"],
                "article_authors": context["authors"],
                "article_type": context["article_type"],
                "section": context["section"],
                "paragraph": context["text"],
                "abstract": abstracts[context.get("article_id")],
                "pubmed_id": context["pubmed_id"],
            }
        )
        answer["metadata"] = metadata

    del answer["paragraphs"]
    return answer



"""Utilities for the master API."""

import logging
import re
from pathlib import Path

logger = logging.getLogger(__name__)


def format_issn(source_issns: str | None) -> str | None:
    """Reformat issns to contain dash and add heading zeroes.

    Parameters
    ----------
    source_issns
        Single str containing original issns from the index.

    Returns
    -------
    str | None
        Single str containing formatted issns or None.
    """
    issn_pattern = r"^\d{4}-\d{3}[0-9X]$"

    if source_issns is None:
        return None
    else:
        issns = []

        for issn in source_issns.split():
            issn = issn.rjust(8, "0")

            issn = issn[:4] + "-" + issn[4:]
            if not re.match(issn_pattern, issn):
                raise ValueError(f"ISSN '{issn}' not in correct format.")
            issns.append(issn)

        return " ".join(issns)


def find_files(
    input_path: Path,
    recursive: bool,
    match_filename: str | None = None,
) -> list[Path]:
    """Find files inside of `input_path`.

    Parameters
    ----------
    input_path
        File or directory to consider.
    recursive
        If True, directories and all subdirectories are considered in a recursive way.
    match_filename
        Only filename matching match_filename are kept.

    Returns
    -------
    inputs : list[Path]
        List of kept files.

    Raises
    ------
    ValueError
        If the input_path does not exists.
    """
    if input_path.is_file():
        return [input_path]

    elif input_path.is_dir():
        if recursive:
            pattern = "**/*"
        else:
            pattern = "*"
        files = (x for x in input_path.glob(pattern) if x.is_file())

        if match_filename is None:
            selected = files
        elif match_filename == "":
            raise ValueError("Value for argument 'match-filename' should not be empty!")
        else:
            regex = re.compile(match_filename)
            selected = (x for x in files if regex.fullmatch(x.name))

        return sorted(selected)

    else:
        raise ValueError(
            "Argument 'input_path' should be a path to an existing file or directory!"
        )



"""Script to create, delete or reset an index in a document store."""

from __future__ import annotations

import argparse
import logging
import os
from typing import Any

from scholarag.document_stores import BaseSearch, ElasticSearch, OpenSearch
from scholarag.document_stores.elastic import (
    MAPPINGS_PARAGRAPHS as ELASTICSEARCH_MAPPINGS_PARAGRAPHS,
)
from scholarag.document_stores.elastic import SETTINGS as ELASTICSEARCH_SETTINGS
from scholarag.document_stores.open import (
    MAPPINGS_PARAGRAPHS as OPENSEARCH_MAPPINGS_PARAGRAPHS,
)
from scholarag.document_stores.open import SETTINGS as OPENSEARCH_SETTINGS

logger = logging.getLogger("create_index")
logging.getLogger("elastic_transport").setLevel(logging.WARNING)


def get_parser() -> argparse.ArgumentParser:
    """Get parser for command line arguments."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "task",
        type=str,
        default="create",
        choices=["create", "delete", "reset"],
        help="Specify the action to take. Reset = delete -> create.",
    )
    parser.add_argument(
        "index",
        type=str,
        help="Name of the index to deal with.",
    )
    parser.add_argument(
        "db_url",
        type=str,
        help="URL of the database instance. Format: HOST:PORT",
    )
    parser.add_argument(
        "--db-type",
        type=str,
        choices=["elasticsearch", "opensearch"],
        default="opensearch",
        help="Type of database.",
    )
    parser.add_argument(
        "--user",
        type=str,
        default=None,
        help=(
            "User of the database instance. The password is read from the environment"
            " under SCHOLARAG__DB__PASSWORD."
        ),
    )
    parser.add_argument(
        "--n-shards", default=2, type=int, help="Number of shards for the index."
    )
    parser.add_argument(
        "--n-replicas", default=1, type=int, help="Number of replicas for the index."
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Control verbosity",
    )
    return parser


def manage_index(
    task: str,
    document_store: BaseSearch,
    index: str,
    mappings: dict[str, Any] | None = None,
    settings: dict[str, str] | None = None,
) -> None:
    """Run the logic."""
    if task in ["delete", "reset"]:
        document_store.remove_index(index)
        logger.info(f"Successfully removed index {index}")
    if task in ["create", "reset"] and mappings is not None and settings is not None:
        document_store.create_index(index, settings, mappings)
        logger.info(f"Successfully created index {index}")


def main() -> int:
    """Run the main logic."""
    parser = get_parser()
    args = parser.parse_args()

    logging_level = logging.INFO if args.verbose else logging.WARNING

    # setup logging
    logging.basicConfig(
        format="[%(levelname)s]  %(asctime)s %(name)s  %(message)s", level=logging_level
    )

    host, _, port = args.db_url.rpartition(":")
    password = os.getenv("SCHOLARAG__DB__PASSWORD")
    if not password:
        logger.warning("SCHOLARAG__DB__PASSWORD variable not set. defaulting to None.")

    if args.db_type == "elasticsearch":
        document_store = (
            ElasticSearch(
                host=host,
                port=port,
                user=args.user,
                password=password,
                use_ssl_and_verify_certs=False,
            )  # type: ignore
            if args.user is not None
            else ElasticSearch(host=host, port=port, use_ssl_and_verify_certs=False)  # type: ignore
        )
        mappings = ELASTICSEARCH_MAPPINGS_PARAGRAPHS
        settings = ELASTICSEARCH_SETTINGS

    elif args.db_type == "opensearch":
        # Opensearch doesn't accept the http protocol at the beginning of the host.
        # While we could rely on the user to not provide it, any new user of our code
        # would get highly confused. So I handle it here.
        document_store = (
            OpenSearch(
                host=host,
                port=port,
                user=args.user,
                password=password,
                use_ssl_and_verify_certs=False,  # type: ignore
            )
            if args.user is not None
            else OpenSearch(host=host, port=port, use_ssl_and_verify_certs=False)  # type: ignore
        )
        mappings = OPENSEARCH_MAPPINGS_PARAGRAPHS
        settings = OPENSEARCH_SETTINGS
    else:
        raise ValueError(
            "The db type should be either 'elasticsearch' or 'opensearch'."
        )

    settings["number_of_shards"] = args.n_shards
    settings["number_of_replicas"] = args.n_replicas

    manage_index(
        task=args.task,
        document_store=document_store,
        index=args.index,
        mappings=mappings,
        settings=settings,
    )

    raise SystemExit(0)


if __name__ == "__main__":
    main()



"""Start the Scholarag API."""

import argparse
from pathlib import Path

import uvicorn


def get_parser() -> argparse.ArgumentParser:
    """Get parser for command line arguments."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "--host",
        type=str,
        default="127.0.0.1",
        help="Host used by the app.",
    )
    parser.add_argument(
        "--port",
        type=int,
        default=8000,
        help="Port used by the app",
    )
    parser.add_argument(
        "--env",
        type=Path,
        default=None,
        help=(
            "Path to the env file for app config. See example at"
            " https://github.com/BlueBrain/scholarag/.env.example. Reads from local"
            " '.env' file in cwd by default."
        ),
    )
    parser.add_argument(
        "--workers",
        type=int,
        default=1,
        help="Number of workers used by the app.",
    )
    return parser


def main() -> None:
    """Run main logic."""
    parser = get_parser()
    args = parser.parse_args()
    uvicorn.run(
        "scholarag.app.main:app",
        host=args.host,
        port=args.port,
        workers=args.workers,
        env_file=args.env,
    )


if __name__ == "__main__":
    main()



"""Class to use the Cohere reranking service."""

import logging
import time
from typing import Any

import cohere
from pydantic import BaseModel, ConfigDict, model_validator

logger = logging.getLogger(__name__)


class CohereRerankingService(BaseModel):
    """Main class to call Cohere's reranking endpoint."""

    api_key: str | None = None
    client: cohere.Client
    async_client: cohere.AsyncClient
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @model_validator(mode="before")
    def create_client(cls, values: dict[str, Any]) -> dict[str, Any]:
        """Instantiate the clients upon class creation."""
        if values.get("api_key") is not None:
            client = cohere.Client(values["api_key"])
            async_client = cohere.AsyncClient(values["api_key"])
        else:
            raise ValueError(
                "No API key provided. You must provide a Cohere API key to call methods"
                " of this class."
            )
        values["client"] = client
        values["async_client"] = async_client
        return values

    def run(
        self, query: str, contexts: list[str]
    ) -> list[dict[str, str | float | int | None]]:
        """Return contexts in decreasing likeliness order of holding the answer to the question.

        Parameters
        ----------
        query
            Question to answer.
        contexts
            Contexts to rerank.

        Returns
        -------
        response: list[dict[str, str | float | int | None]]
            List of dictionaries containing in score order the text and the associated score.
        """
        response = self.client.rerank(
            model="rerank-english-v2.0",
            query=query,
            documents=contexts,
            return_documents=True,
            top_n=len(contexts),
            max_chunks_per_doc=10_000 // len(contexts),
        )
        reranked_contexts = [
            {
                "text": (
                    reranked.document.text if reranked.document is not None else None
                ),
                "score": reranked.relevance_score,
                "index": reranked.index,
            }
            for reranked in response.results
        ]
        return reranked_contexts

    async def arun(
        self, query: str, contexts: list[str]
    ) -> list[dict[str, str | float | int | None]]:
        """Return contexts in decreasing likeliness order of holding the answer to the question.

        Parameters
        ----------
        query
            Question to answer.
        contexts
            Contexts to rerank.

        Returns
        -------
        response: list[dict[str, str | float | int | None]]
            List of dictionaries containing in score order the text and the associated score.
        """
        response = await self.async_client.rerank(
            model="rerank-english-v2.0",
            query=query,
            documents=contexts,
            return_documents=True,
            top_n=len(contexts),
            max_chunks_per_doc=10_000 // len(contexts),
        )

        reranked_contexts = [
            {
                "text": (
                    reranked.document.text if reranked.document is not None else None
                ),
                "score": reranked.relevance_score,
                "index": reranked.index,
            }
            for reranked in response.results
        ]
        return reranked_contexts

    async def rerank(
        self,
        query: str,
        contexts: list[dict[str, Any]],
        reranker_k: int,
    ) -> tuple[list[dict[str, Any]], list[str], tuple[float], tuple[int]]:
        """Run the reranking logic.

        Parameters
        ----------
        query
            Query to use to handle the reranking of the contexts.
        contexts
            Paragraphs saved under dictionaries with metadata to rerank in regard of the query.
        reranker_k
            Number of paragraphs to keep after the reranking.

        Returns
        -------
        new_contexts: list[dict[str, Any]]
            Contexts kept and reordered after the reranking.
        new_contexts_text: list[str]
            Contexts text after reranking.
        scores: tuple[float]
            Scores given by reranker for the different contexts
        indices: tuple[int]
            Initial indices of the contexts before the reranking phase.
        """
        contexts_text = [context["text"] for context in contexts]
        logger.info(
            f"Reranking {len(contexts_text)} contexts with Cohere reranker. "
            f"Keeping the {reranker_k} most relevant ones..."
        )
        start = time.time()
        reranked = await self.arun(query=query, contexts=contexts_text)
        logger.info(f"Reranking took {time.time() - start}s.")
        reranked_contexts, scores, indices = zip(
            *[
                (context["text"], context["score"], context["index"])
                for context in reranked[:reranker_k]
            ]
        )
        logger.info(f"New contexts position: {indices} with scores: {scores}.")
        new_contexts = [contexts[i] for i in indices]
        new_contexts_text = list(reranked_contexts)
        return new_contexts, new_contexts_text, scores, indices



"""FastAPI for the Scholarag SDK."""

import logging
from contextlib import asynccontextmanager
from logging.config import dictConfig
from typing import Annotated, Any, AsyncContextManager
from uuid import uuid4

import sentry_sdk
from asgi_correlation_id import CorrelationIdMiddleware
from fastapi import Depends, FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi_pagination import add_pagination
from fastapi_pagination.utils import disable_installed_extensions_check
from starlette.middleware.base import BaseHTTPMiddleware

from scholarag import __version__
from scholarag.app.config import Settings
from scholarag.app.dependencies import get_settings
from scholarag.app.middleware import get_and_set_cache, get_cache, strip_path_prefix
from scholarag.app.routers import qa, retrieval, suggestions

stry = sentry_sdk.init(
    release=__version__,
    traces_sample_rate=1.0,
    profiles_sample_rate=1.0,
)

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "filters": {
        "correlation_id": {
            "()": "asgi_correlation_id.CorrelationIdFilter",
            "uuid_length": 32,
            "default_value": "-",
        },
    },
    "formatters": {
        "request_id": {
            "class": "logging.Formatter",
            "format": (
                "[%(levelname)s] %(asctime)s (%(correlation_id)s) %(name)s %(message)s"
            ),
        },
    },
    "handlers": {
        "request_id": {
            "class": "logging.StreamHandler",
            "filters": ["correlation_id"],
            "formatter": "request_id",
        },
    },
    "loggers": {
        "": {
            "handlers": ["request_id"],
            "level": "INFO",
            "propagate": True,
        },
    },
}
dictConfig(LOGGING)

logger = logging.getLogger(__name__)


@asynccontextmanager  # type: ignore
async def lifespan(fastapi_app: FastAPI) -> AsyncContextManager[None]:  # type: ignore
    """Read environment (settings of the application)."""
    # hacky but works: https://github.com/tiangolo/fastapi/issues/425
    app_settings = fastapi_app.dependency_overrides.get(get_settings, get_settings)()

    prefix = app_settings.misc.application_prefix
    fastapi_app.openapi_url = f"{prefix}/openapi.json"
    fastapi_app.servers = [{"url": prefix}]
    # Do not rely on the middleware order in the list "fastapi_app.user_middleware" since this is subject to changes.
    try:
        cors_middleware = filter(
            lambda x: x.__dict__["cls"] == CORSMiddleware, fastapi_app.user_middleware
        ).__next__()
        cors_middleware.kwargs["allow_origins"] = (
            app_settings.misc.cors_origins.replace(" ", "").split(",")
        )
    except StopIteration:
        pass

    get_cache(app_settings)

    logging.getLogger().setLevel(app_settings.logging.external_packages.upper())
    logging.getLogger("scholarag").setLevel(app_settings.logging.level.upper())
    logging.getLogger("app").setLevel(app_settings.logging.level.upper())

    yield


app = FastAPI(
    title="Literature search",
    summary="API for answering scientific questions",
    version=__version__,
    swagger_ui_parameters={"tryItOutEnabled": True},
    lifespan=lifespan,
)

app.include_router(suggestions.router)
app.include_router(qa.router)
app.include_router(retrieval.router)
app.add_middleware(BaseHTTPMiddleware, dispatch=get_and_set_cache)
app.add_middleware(
    CORSMiddleware,
    allow_origin_regex=r"^http:\/\/localhost:(\d+)\/?.*$",
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)
app.add_middleware(
    CorrelationIdMiddleware,
    header_name="X-Request-ID",
    update_request_header=True,
    generator=lambda: uuid4().hex,
    transformer=lambda a: a,
)
app.add_middleware(BaseHTTPMiddleware, dispatch=strip_path_prefix)

# fastapi_pagination settings
disable_installed_extensions_check()
add_pagination(app)


@app.get("/healthz")
def healthz() -> str:
    """Check the health of the API."""
    return "200"


@app.get("/")
def readyz() -> dict[str, str]:
    """Check if the API is ready to accept traffic."""
    return {"status": "ok"}


@app.get("/settings")
def settings(settings: Annotated[Settings, Depends(get_settings)]) -> Any:
    """Show complete settings of the backend.

    Did not add return model since it pollutes the Swagger UI.
    """
    return settings



import random
from pathlib import Path

import pytest
from scholarag.ds_utils import check_docs_exists_in_db, ds_upload, setup_parsing_ds
from scholarag.services import ParsingService


@pytest.mark.asyncio
async def test_setup_parsing_ds(get_testing_async_ds_client):
    ds_client, _ = get_testing_async_ds_client
    db_url, db_type = (
        ("http://localhost:9201", "elasticsearch")
        if "ElasticSearch" in ds_client.__class__.__name__
        else ("localhost:9200", "opensearch")
    )
    ds_client, parsing_client = await setup_parsing_ds(
        db_url=db_url,
        db_type=db_type,
    )
    assert await ds_client.client.ping()
    assert isinstance(parsing_client, ParsingService)


@pytest.mark.asyncio
async def test_ds_upload_and_check_in_db(get_testing_async_ds_client):
    ds_client, parameters = get_testing_async_ds_client
    filenames = [Path("file1.xml"), Path("file2.xml"), Path("file3.xml")]
    index = "paragraphs_ds_upload"
    results = [
        {
            "uid": "uid1",
            "abstract": ["This is a first paragraph", "This is a second"],
            "authors": ["author a", "author b"],
            "title": "This is a fake title",
            "pubmed_id": "12345",
            "pmc_id": "PMC123456",
            "arxiv_id": "arxiv_id1",
            "doi": "DOI12345",
            "date": "2020-12-21",
            "journal": "Journal name",
            "article_type": "research_article",
            "section_paragraphs": [
                ("Introduction", "This is the introduction"),
                ("Conclusion", "This is the conclusion"),
            ],
        },
        None,
        {
            "uid": "uid1",
            "abstract": ["This is a first great paragraph", "This is a bad second"],
            "authors": ["author a", "author b"],
            "title": "This is a fake title",
            "pubmed_id": "12345",
            "pmc_id": "PMC654321",
            "arxiv_id": "arxiv_id1",
            "doi": "DOI12345",
            "date": "2020-12-21",
            "journal": "Journal name",
            "article_type": "research_article",
            "section_paragraphs": [
                ("Introduction", "This is the second introduction"),
                ("Conclusion", "This is the second conclusion"),
            ],
        },
    ]

    await ds_client.create_index(
        index=index,
        mappings=parameters[0],
        settings=parameters[1],
    )

    files_failing = await ds_upload(
        filenames=filenames,
        results=results,
        indices=[index for _ in range(len(filenames))],
        ds_client=ds_client,
    )

    await ds_client.client.indices.refresh()
    assert await ds_client.count_documents(index) == 8

    assert files_failing[0] == Path("file2.xml")

    # Upload five more paragraphs using length lower bound
    results[0]["uid"] = str(random.random())
    results[2]["uid"] = str(random.random())
    await ds_upload(
        filenames=filenames,
        results=results,
        indices=[index for _ in range(len(filenames))],
        ds_client=ds_client,
        min_paragraphs_length=23,
    )

    await ds_client.client.indices.refresh()
    assert await ds_client.count_documents(index) == 13

    # Upload three more paragraphs using length upper bound
    results[0]["uid"] = str(random.random())
    results[2]["uid"] = str(random.random())
    await ds_upload(
        filenames=filenames,
        results=results,
        indices=[index for _ in range(len(filenames))],
        ds_client=ds_client,
        max_paragraphs_length=23,
    )

    await ds_client.client.indices.refresh()
    assert await ds_client.count_documents(index) == 16

    # Try to upload similar paragraphs within the same article
    results[0]["uid"] = str(random.random())
    results[2] = results[0]
    await ds_upload(
        filenames=filenames,
        results=results,
        indices=[index for _ in range(len(filenames))],
        ds_client=ds_client,
    )

    await ds_client.client.indices.refresh()
    assert await ds_client.count_documents(index) == 20


@pytest.mark.parametrize(
    "pmc_ids, expected_existing_ids",
    [
        (
            ["PMC123456", "PMC654321", "PMC135246"],
            ["PMC123456", "PMC654321", "PMC135246"],
        ),
        (
            ["PMC123456", "PMC654321", "PMC678438"],
            [
                "PMC123456",
                "PMC654321",
            ],
        ),
        (
            [
                "PMC349827",
                "PMC049123",
            ],
            [],
        ),
    ],
)
@pytest.mark.asyncio
async def test_check_docs_exists_in_db(
    pmc_ids, expected_existing_ids, get_testing_async_ds_client
):
    ds_client, parameters = get_testing_async_ds_client
    index = "check_docs_in_db"

    await ds_client.create_index(
        index=index,
        mappings=parameters[0],
        settings=parameters[1],
    )

    doc_1 = {
        "_index": index,
        "_id": 1,
        "_source": {
            "text": "test of an amazing function",
            "title": "test_article",
            "paragraph_id": "1",
            "article_id": "article_id",
            "journal": "8765-4321",
            "pmc_id": "PMC123456",
        },
    }

    doc_2 = {
        "_index": index,
        "_id": 2,
        "_source": {
            "text": "The bird sings very loudly",
            "title": "test_article",
            "paragraph_id": "1",
            "article_id": "article_id",
            "journal": None,
            "pmc_id": "PMC654321",
        },
    }

    doc_3 = {
        "_index": index,
        "_id": 3,
        "_source": {
            "text": "This document is a bad test, I don't want to retrieve it",
            "title": "test_article",
            "paragraph_id": "1",
            "article_id": "article_id",
            "journal": "1234-5678",
            "pmc_id": "PMC135246",
        },
    }
    doc_bulk = [doc_1, doc_2, doc_3]

    await ds_client.bulk(doc_bulk)
    await ds_client.client.indices.refresh()

    docs = await ds_client.search(index, query={"match_all": {}})
    assert len(docs["hits"]["hits"]) == 3

    existings_ids = await check_docs_exists_in_db(ds_client, index, pmc_ids)
    assert set(existings_ids) == set(expected_existing_ids)



"""Script to create impact factor index."""

from __future__ import annotations

import argparse
import logging
import os
from pathlib import Path
from typing import Any

import opensearchpy.exceptions
import pandas as pd

from scholarag.document_stores import BaseSearch, ElasticSearch, OpenSearch
from scholarag.document_stores.elastic import SETTINGS as ELASTICSEARCH_SETTINGS
from scholarag.document_stores.open import SETTINGS as OPENSEARCH_SETTINGS

logger = logging.getLogger("create_impact_factor_index")
logging.getLogger("elastic_transport").setLevel(logging.WARNING)
logging.getLogger("opensearch").setLevel(logging.WARNING)

IMPACT_FACTORS_MAPPING = {
    "dynamic": "strict",
    "properties": {
        "Citation Count": {"type": "integer"},
        "CiteScore": {"type": "float"},
        "E-ISSN": {"type": "keyword"},
        "Main Publisher": {"type": "keyword"},
        "Open Access": {"type": "keyword"},
        "Percent Cited": {"type": "float"},
        "Percentile": {"type": "integer"},
        "Print ISSN": {"type": "keyword"},
        "Publisher": {"type": "keyword"},
        "Quartile": {"type": "integer"},
        "RANK": {"type": "integer"},
        "Rank Out Of": {"type": "text"},
        "SJR": {"type": "float"},
        "SNIP": {"type": "float"},
        "Scholarly Output": {"type": "integer"},
        "Scopus ASJC Code (Sub-subject Area)": {"type": "integer"},
        "Scopus Source ID": {"type": "keyword"},
        "Scopus Sub-Subject Area": {"type": "text"},
        "Title": {"type": "text"},
        "Top 10% (CiteScore Percentile)": {"type": "keyword"},
        "Type": {"type": "keyword"},
        "URL Scopus Source ID": {"type": "text"},
    },
}


def get_parser() -> argparse.ArgumentParser:
    """Get parser for command line arguments."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "from_file",
        type=Path,
        help="Path to file containing impact factor information.",
    )
    parser.add_argument(
        "index",
        type=str,
        help="Name of the index to create.",
    )
    parser.add_argument(
        "db_url",
        type=str,
        help="URL of the database instance. Format: HOST:PORT",
    )
    parser.add_argument(
        "--db-type",
        type=str,
        choices=["elasticsearch", "opensearch"],
        default="opensearch",
        help="Type of database.",
    )
    parser.add_argument(
        "--user",
        type=str,
        default=None,
        help=(
            "User of the database instance. The password is read from the environment"
            " under SCHOLARAG__DB__PASSWORD."
        ),
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Control verbosity",
    )
    return parser


def run(
    document_store: BaseSearch,
    from_file: Path,
    index: str,
    settings: dict[str, Any] | None = None,
) -> None:
    """Create and populate the impact factors index from a file.

    Parameters
    ----------
    document_store
        Document store to populate with the new impact factors index.
    from_file
        File containing the data to populate the index.
    index
        Name of the index to create in the document store.
    settings
        Settings to use when creating the index.
    """
    if not from_file.exists():
        raise ValueError(f"The file {from_file} does not exist.")

    if index in document_store.get_available_indexes():
        raise ValueError(f"The index {index} already exists.")

    document_store.create_index(
        index,
        mappings=IMPACT_FACTORS_MAPPING,
        settings=settings,
    )

    document_store.client.indices.refresh()  # type: ignore

    df = pd.read_excel(from_file)
    # Clean the DataFrame, remove rows with no E-ISSN, fill NaN for others
    df = df[df["E-ISSN"].notna()]
    df = df.fillna(
        {
            "SNIP": 0,
            "SJR": 0,
            "Publisher": "",
            "Main Publisher": "",
            "Print ISSN": "",
        },
    )

    data = df.to_dict("records")
    successful, failed = [], []
    for d in data:
        # Searching the impact factors through the E-ISSN in the app, so making it the ID
        if document_store.exists(index, str(d["E-ISSN"])):
            continue
        logger.debug(f"Uploading {d['E-ISSN']}")
        try:
            document_store.add_document(
                index=index,
                doc=d,
                doc_id=str(d["E-ISSN"]),
            )
            successful.append(d["E-ISSN"])
        except (opensearchpy.exceptions.RequestError, RuntimeError):
            logger.warning(f"Failed to upload {d['E-ISSN']}")
            failed.append(d["E-ISSN"])

    logger.debug(
        f"{len(successful)} docs were successfully uploaded, {len(failed)} failed."
    )


def main() -> int:
    """Run the main logic."""
    parser = get_parser()
    args = parser.parse_args()

    logging_level = logging.INFO if args.verbose else logging.WARNING

    # setup logging
    logging.basicConfig(
        format="[%(levelname)s]  %(asctime)s %(name)s  %(message)s", level=logging_level
    )

    host, _, port = args.db_url.rpartition(":")
    password = os.getenv("SCHOLARAG__DB__PASSWORD")
    if not password:
        logger.warning("SCHOLARAG__DB__PASSWORD variable not set. defaulting to None.")

    if args.db_type == "elasticsearch":
        document_store = (
            ElasticSearch(
                host=host,
                port=port,
                user=args.user,
                password=password,
                use_ssl_and_verify_certs=False,
            )  # type: ignore
            if args.user is not None
            else ElasticSearch(host=host, port=port, use_ssl_and_verify_certs=False)  # type: ignore
        )
        settings = ELASTICSEARCH_SETTINGS

    elif args.db_type == "opensearch":
        # Opensearch doesn't accept the http protocol at the beginning of the host.
        # While we could rely on the user to not provide it, any new user of our code
        # would get highly confused. So I handle it here.
        document_store = (
            OpenSearch(
                host=host,
                port=port,
                user=args.user,
                password=password,
                use_ssl_and_verify_certs=False,  # type: ignore
            )
            if args.user is not None
            else OpenSearch(host=host, port=port, use_ssl_and_verify_certs=False)  # type: ignore
        )
        settings = OPENSEARCH_SETTINGS
    else:
        raise ValueError(
            "The db type should be either 'elasticsearch' or 'opensearch'."
        )

    run(
        document_store=document_store,
        from_file=args.from_file,
        index=args.index,
        settings=settings,
    )

    return 0


if __name__ == "__main__":
    main()



"""Endpoints for question answering pipeline."""

import logging
import re
import time
from typing import Annotated, Any

from cohere.errors import TooManyRequestsError
from fastapi import APIRouter, Depends, HTTPException
from fastapi.responses import StreamingResponse
from httpx import AsyncClient
from openai import AsyncOpenAI, BadRequestError

from scholarag.app.config import Settings
from scholarag.app.dependencies import (
    ErrorCode,
    get_ds_client,
    get_generative_qas,
    get_httpx_client,
    get_openai_client,
    get_query_from_params,
    get_reranker,
    get_rts,
    get_settings,
)
from scholarag.app.schemas import (
    GenerativeQARequest,
    GenerativeQAResponse,
    ParagraphMetadata,
    PassthroughRequest,
    PassthroughResponse,
)
from scholarag.app.streaming import stream_response
from scholarag.document_stores import AsyncBaseSearch
from scholarag.generative_question_answering import GenerativeQAWithSources
from scholarag.retrieve_metadata import MetaDataRetriever
from scholarag.services import CohereRerankingService, RetrievalService

router = APIRouter(prefix="/qa", tags=["Question answering"])

logger = logging.getLogger(__name__)


@router.post(
    "/passthrough",
    response_model=PassthroughResponse,
    responses={
        500: {
            "description": (
                f"If code={ErrorCode.INCOMPLETE_ANSWER.value}, max number of token"
                " reached during the answer of generative model."
            ),
            "content": {
                "application/json": {
                    "schema": {
                        "example": {
                            "detail": {
                                "code": 1,
                                "detail": "Message",
                            }
                        }
                    }
                }
            },
        },
        413: {
            "description": (
                f"If code={ErrorCode.INPUT_EXCEEDED_MAX_TOKENS.value}, max number of"
                " token reached for the generative model."
            ),
            "content": {
                "application/json": {
                    "schema": {
                        "example": {
                            "detail": {
                                "code": 1,
                                "detail": "Message",
                            }
                        }
                    }
                }
            },
        },
    },
)
async def passthrough(
    request: PassthroughRequest,
    qas: Annotated[AsyncOpenAI, Depends(get_openai_client)],
    settings: Annotated[Settings, Depends(get_settings)],
) -> PassthroughResponse:
    """Answer the query given the context with a generative model.
    \f

    Parameters
    ----------
    request
        Body of the request sent to the server.
    qas
        Client to the service processing the query
    settings
        Global settings of the application

    Returns
    -------
        A single answer.
    """  # noqa: D301, D400, D205
    start = time.time()
    logger.info("Finding answers ...")
    try:
        answer = await qas.chat.completions.create(
            messages=[
                {
                    "role": "user",
                    "content": request.query,
                }
            ],
            model=settings.generative.openai.model,
        )
    except BadRequestError as err:
        numbers = re.findall(r"\d+", err.message)[1:]
        if len(numbers) == 2:
            message = (
                f"This model's maximum context length is {numbers[0]} tokens. However, "
                f"your messages resulted in {numbers[1]} tokens. Please reduce"
                " retriever_k."
            )
            raise HTTPException(
                status_code=413,
                detail={
                    "code": ErrorCode.INPUT_EXCEEDED_MAX_TOKENS.value,
                    "detail": message,
                },
            )
        else:
            raise HTTPException(
                status_code=413,
                detail={
                    "code": ErrorCode.INPUT_EXCEEDED_MAX_TOKENS.value,
                    "detail": "OpenAI error.",
                },
            )

    if answer.choices[0].finish_reason == "length":
        raise HTTPException(
            status_code=500,
            detail={
                "code": ErrorCode.INCOMPLETE_ANSWER.value,
                "detail": (
                    "The LLM did not have enough completion tokens available to finish"
                    " its answer."
                ),
                "raw_answer": answer.choices[0].message.content,
            },
        )
    logger.info(f"Found answer in {time.time() - start}")
    return PassthroughResponse(answer=answer.choices[0].message.content)


@router.post(
    "/generative",
    response_model=GenerativeQAResponse,
    responses={
        500: {
            "description": (
                f"If code={ErrorCode.NO_DB_ENTRIES.value}, no relevant context is found"
                f" in the database. If code={ErrorCode.NO_ANSWER_FOUND.value}, no"
                " answer is found by the LLM from the extracted contexts. If"
                f" code={ErrorCode.INCOMPLETE_ANSWER.value}, max number of token"
                " reached during the answer of generative model -> Please reduce"
                " retriever_k of 1 or 2 paragraphs. If"
                f" code={ErrorCode.MAX_COHERE_REQUESTS.value}, max number of requests"
                " reached for the reranking model -> Wait a bit or turn it off."
            ),
            "content": {
                "application/json": {
                    "schema": {
                        "example": {
                            "detail": {
                                "code": 1,
                                "detail": "Message",
                            }
                        }
                    }
                }
            },
        },
        413: {
            "description": (
                f"If code={ErrorCode.INPUT_EXCEEDED_MAX_TOKENS.value}, max number of"
                " token reached for the generative model -> Reduce retriever_k."
            ),
            "content": {
                "application/json": {
                    "schema": {
                        "example": {
                            "detail": {
                                "code": 1,
                                "detail": "Message",
                            }
                        }
                    }
                }
            },
        },
    },
)
async def generative_qa(
    request: GenerativeQARequest,
    rts: Annotated[RetrievalService, Depends(get_rts)],
    qas: Annotated[GenerativeQAWithSources, Depends(get_generative_qas)],
    rs: Annotated[CohereRerankingService | None, Depends(get_reranker)],
    ds_client: Annotated[AsyncBaseSearch, Depends(get_ds_client)],
    httpx_client: Annotated[AsyncClient, Depends(get_httpx_client)],
    filter_query: Annotated[dict[str, Any], Depends(get_query_from_params)],
    settings: Annotated[Settings, Depends(get_settings)],
) -> GenerativeQAResponse:
    """Answer the query given the context with a generative model.
    \f

    Parameters
    ----------
    request
        Body of the request sent to the server.
    rts
        Retrieval service.
    qas
        Question_answering service to generate the answer based on the retrieved documents.
    rs
        Reranker service.
    ds_client
        Document store client.
    httpx_client
        Httpx async client to send external requests.
    filter_query
        Filtering query for paragraph retrieval.
    settings
        Global settings of the application

    Returns
    -------
        A single answer with metadata of each relevant source.
    """  # noqa: D301, D400, D205
    start = time.time()
    logger.info("Finding answers ...")

    contexts = await rts.arun(
        ds_client=ds_client,
        query=request.query,
        retriever_k=request.retriever_k,
        db_filter=filter_query,
    )
    if not contexts:
        raise HTTPException(
            status_code=500,
            detail={
                "code": ErrorCode.NO_DB_ENTRIES.value,
                "detail": (
                    "No document found. Modify the filters or the query and try again."
                ),
            },
        )
    logger.info(f"Semantic search took {time.time() - start}s")

    if rs is None or not request.use_reranker:
        contexts_text = [context["text"] for context in contexts]
        indices = tuple(range(len(contexts_text)))
        scores = None

    else:
        try:
            contexts, contexts_text, scores, indices = await rs.rerank(
                query=request.query,
                contexts=contexts,
                reranker_k=request.reranker_k,
            )
        except TooManyRequestsError:
            raise HTTPException(
                status_code=500,
                detail={
                    "code": ErrorCode.MAX_COHERE_REQUESTS.value,
                    "detail": (
                        "Max number of requests reached for Cohere reranker. Wait a"
                        " little bit and try again, or disable the reranker."
                    ),
                },
            )

    metadata_retriever = MetaDataRetriever(
        external_apis=settings.metadata.external_apis
    )
    metadata_retriever.schedule_bulk_request(
        qas.arun,  # type: ignore
        **{
            "query": request.query,
            "contexts": contexts_text,
            "prompt_template": settings.generative.prompt_template.get_secret_value(),
        },
    )
    try:
        fetched_metadata = await metadata_retriever.retrieve_metadata(
            contexts,
            ds_client,
            settings.db.index_journals,
            settings.db.index_paragraphs,
            httpx_client,
        )
    except BadRequestError as err:
        raise HTTPException(
            status_code=413,
            detail={
                "code": ErrorCode.INPUT_EXCEEDED_MAX_TOKENS.value,
                "detail": err.body["message"],  # type: ignore
            },
        )

    answer, finish_reason = fetched_metadata["arun"]

    if finish_reason == "length":
        raise HTTPException(
            status_code=500,
            detail={
                "code": ErrorCode.INCOMPLETE_ANSWER.value,
                "detail": (
                    "The LLM did not have enough completion tokens available to finish"
                    " its answer. Please decrease the reranker_k or retriever_k value"
                    " by 1 or 2 depending of whether you are using the reranker or not."
                ),
                "raw_answer": answer["raw_answer"],
            },
        )

    if answer["paragraphs"] is None or len(answer["paragraphs"]) == 0:
        raise HTTPException(
            status_code=500,
            detail={
                "code": ErrorCode.NO_ANSWER_FOUND.value,
                "detail": "The LLM did not provide any source to answer the question.",
                "raw_answer": answer["raw_answer"],
            },
        )
    else:
        citedby_counts = fetched_metadata.get("get_citation_count", {})
        journal_names = fetched_metadata.get("get_journal_name", {})
        impact_factors = fetched_metadata["get_impact_factors"]
        abstracts = fetched_metadata["recreate_abstract"]

        context_ids: list[int] = answer["paragraphs"]
        logger.info("Adding article metadata to the answers")
        metadata = []
        for context_id in context_ids:
            context = contexts[context_id]

            metadata.append(
                ParagraphMetadata(
                    **{
                        "impact_factor": impact_factors[context.get("journal")],
                        "journal_name": journal_names.get(context.get("journal")),
                        "journal_issn": context.get("journal"),
                        "date": context.get("date"),
                        "cited_by": citedby_counts.get(context.get("doi")),
                        "article_id": context["article_id"],
                        "ds_document_id": str(context["document_id"]),
                        "context_id": indices[context_id],
                        "reranking_score": (
                            scores[context_id] if scores is not None else None
                        ),
                        "article_doi": context["doi"],
                        "article_title": context["title"],
                        "article_authors": context["authors"],
                        "article_type": context["article_type"],
                        "section": context["section"],
                        "paragraph": context["text"],
                        "abstract": abstracts[context.get("article_id")],
                        "pubmed_id": context["pubmed_id"],
                    }
                )
            )
        answer["metadata"] = metadata

    del answer["paragraphs"]
    logger.info(f"Total time to generate a complete answer: {time.time() - start}")
    return GenerativeQAResponse(**answer)


@router.post(
    "/streamed_generative",
    response_model=GenerativeQAResponse,
    responses={
        500: {
            "description": (
                f"If code={ErrorCode.NO_DB_ENTRIES.value}, no relevant context is found"
                f" in the database. If code={ErrorCode.NO_ANSWER_FOUND.value}, no"
                " answer is found by the LLM from the extracted contexts."
            ),
            "content": {
                "application/json": {
                    "schema": {
                        "example": {
                            "detail": {
                                "code": 1,
                                "detail": "Message",
                            }
                        }
                    }
                }
            },
        },
        413: {
            "description": (
                f"If code={ErrorCode.INPUT_EXCEEDED_MAX_TOKENS.value}, max number of"
                " token reached for the generative model -> Reduce retriever_k."
            ),
            "content": {
                "application/json": {
                    "schema": {
                        "example": {
                            "detail": {
                                "code": 1,
                                "detail": "Message",
                            }
                        }
                    }
                }
            },
        },
    },
)
async def streamed_generative_qa(
    request: GenerativeQARequest,
    rts: Annotated[RetrievalService, Depends(get_rts)],
    qas: Annotated[GenerativeQAWithSources, Depends(get_generative_qas)],
    rs: Annotated[CohereRerankingService | None, Depends(get_reranker)],
    ds_client: Annotated[AsyncBaseSearch, Depends(get_ds_client)],
    filter_query: Annotated[dict[str, Any], Depends(get_query_from_params)],
    settings: Annotated[Settings, Depends(get_settings)],
) -> StreamingResponse:
    """Answer the query given the context with a generative model.
    \f

    Parameters
    ----------
    request
        Body of the request sent to the server.
    rts
        Semantic_search_service for the retrieval.
    qas
        Question_answering service to generate the answer based on the retrieved documents.
    rs
        Reranker service.
    ds_client
        Document store client.
    filter_query
        Filtering query for paragraph retrieval.
    settings
        Global settings of the application

    Returns
    -------
        A single answer with metadata of each relevant source.
    """  # noqa: D301, D400, D205
    start = time.time()
    logger.info("Finding answers ...")

    contexts = await rts.arun(
        ds_client=ds_client,
        query=request.query,
        retriever_k=request.retriever_k,
        db_filter=filter_query,
        max_length=settings.retrieval.max_length,
    )
    if not contexts:
        raise HTTPException(
            status_code=500,
            detail={
                "code": ErrorCode.NO_DB_ENTRIES.value,
                "detail": (
                    "No document found. Modify the filters or the query and try again."
                ),
            },
        )
    logger.info(f"Semantic search took {time.time() - start}s")

    if rs is None or not request.use_reranker:
        contexts_text = [context["text"] for context in contexts]
        indices = tuple(range(len(contexts_text)))
        scores = None
    else:
        contexts, contexts_text, scores, indices = await rs.rerank(
            query=request.query,
            contexts=contexts,
            reranker_k=request.reranker_k,
        )

    metadata_retriever = MetaDataRetriever(
        external_apis=settings.metadata.external_apis
    )

    return StreamingResponse(
        stream_response(
            qas=qas,
            query=request.query,
            contexts_text=contexts_text,
            metadata_retriever=metadata_retriever,
            settings=settings,
            contexts=contexts,
            indices=indices,
            scores=scores,
            ds_client=ds_client,
            index_journals=settings.db.index_journals,
            index_paragraphs=settings.db.index_paragraphs,
        ),
        media_type="text/event-stream",
    )



"""Run the parsing routine."""

from __future__ import annotations

import asyncio
import json
import logging
from contextlib import AsyncExitStack
from pathlib import Path
from typing import Any

from httpx import AsyncClient, AsyncHTTPTransport, Response
from httpx._exceptions import HTTPError
from pydantic import BaseModel

logger = logging.getLogger(__name__)


class ParsingService(BaseModel):
    """Class to parse files.

    Parameters
    ----------
    ignore_errors
        Controls whether the code should raise an exception and stop upon having a request with non 2xx status code.
    max_concurrent_requests
        Maximum number of requests that can be sent simultaneously to the server.
    """

    max_concurrent_requests: int | None = None
    ignore_errors: bool = True

    def run(
        self,
        files: list[Path] | list[bytes],
        url: str,
        multipart_params: dict[str, Any] | None = None,
        timeout: int | None = None,
        httpx_client: AsyncClient | None = None,
    ) -> list[dict[str, Any] | None]:
        """Parse one or multiple files.

        Parameters
        ----------
        files
            Files to parse.
        url
            Url of the parser.
        multipart_params
            Optional request parameters to go along with the file. NOT QUERY PARAMETERS.
        timeout
            Timeout of HTTP requests.
        httpx_client
            Async HTTP Client (Optional)

        Returns
        -------
        parsed: list[dict[str, Any] | None]
            Parsing of the files.
        """
        logger.info(f"Parsing {len(files)} files")
        # create payload
        parsed = asyncio.run(
            self.arun(
                files=files,
                url=url,
                multipart_params=multipart_params,
                timeout=timeout,
                httpx_client=httpx_client,
            )
        )

        return parsed

    async def arun(
        self,
        files: list[Path] | list[bytes],
        url: str,
        multipart_params: dict[str, Any] | None = None,
        timeout: int | None = None,
        httpx_client: AsyncClient | None = None,
    ) -> list[dict[str, str | list[str] | None] | None]:
        """Call async the ETL API.

        Parameters
        ----------
        files
            List of file path or bytes that have to be parsed.
        url
            URL of the target ETL API.
        multipart_params
            Optional parameters to go along with the file. NOT QUERY PARAMETERS. Relevant for pdfs.
        timeout
            Sets global timeout for http requests.
        httpx_client
            Async HTTPx Client.

        Returns
        -------
        output: list[dict[str, str | list[str] | None] | None]
            Output of the requests.
        """
        logger.info(f"Parsing {len(files)} files")
        if self.max_concurrent_requests is not None:
            semaphore = asyncio.Semaphore(self.max_concurrent_requests)
        else:
            semaphore = None

        async with AsyncExitStack() as stack:
            # If the httpx client is None, create it and enter the context manager.
            # This way the client gets closed after sending requests.
            # If the httpx client is provided, we don't close it so that
            # it can be reused. It becomes the user's responsibility to close it.
            if httpx_client is None:
                transport = AsyncHTTPTransport(retries=6)
                httpx_client = AsyncClient(timeout=timeout, transport=transport)
                await stack.enter_async_context(httpx_client)

            tasks = [
                asyncio.create_task(
                    self.asend_request(
                        httpx_client=httpx_client,
                        url=url,
                        file=file,
                        multipart_params=multipart_params,
                        semaphore=semaphore,
                    )
                )
                for file in files
            ]
            res = await asyncio.gather(*tasks, return_exceptions=self.ignore_errors)

        output = [
            (
                response.json()
                if isinstance(response, Response) and response.status_code // 100 == 2
                else None
            )
            for response in res
        ]
        return output

    @staticmethod
    async def asend_request(
        httpx_client: AsyncClient,
        url: str,
        file: Path | bytes,
        multipart_params: dict[str, Any] | None = None,
        semaphore: asyncio.Semaphore | None = None,
    ) -> Response:
        """Call the specified ETL API and returns parsed files.

        Parameters
        ----------
        httpx_client
            Client to use to send the requests.
        url
            URL of the target API.
        file
            Body of the request to send. Should be preferably batched
        multipart_params
            Optional parameters to go along with the file. NOT QUERY PARAMETERS.
        semaphore
            asyncio.Semaphore class used to limit the number of simultaneously outgoing requests.

        Returns
        -------
        result: Response
            Output of the request.
        """
        async with semaphore if semaphore is not None else AsyncExitStack():  # type: ignore
            try:
                # Multipart params for pdfs. Must be "{}" if empty.
                data = {"parameters": json.dumps(multipart_params)}

                if isinstance(file, Path):
                    with file.open("rb") as f:
                        files = [("inp", (file.name, f, "text/xml"))]
                        logger.debug(f"Sending request for file {files[0][1][0]}")
                        result = await httpx_client.post(url, files=files, data=data)
                        logger.debug(
                            f"Received request for file {files[0][1][0]}. response"
                            f" code: {result.status_code}"
                        )
                elif isinstance(file, bytes):
                    logger.debug("Sending bytes contained in the body.")
                    file_bytes = [("inp", ("text.xml", file, "text/xml"))]
                    result = await httpx_client.post(url, files=file_bytes, data=data)
                    logger.debug("Reiceved bytes contained in the body.")
                else:
                    raise ValueError(
                        "Wrong body type for task etl. It should be a pathlib.Path or"
                        " bytes."
                    )

                if result.status_code != 200:
                    file_str = file.decode() if isinstance(file, bytes) else file
                    raise ValueError(
                        f"Something wrong happened for the body {file_str}. "
                        f"The status code is {result.status_code}."
                    )

            except HTTPError as err:
                file_str = file.decode() if isinstance(file, bytes) else file
                raise HTTPError(
                    f"Something wrong happened for the body {file_str}"
                ) from err
        return result



"""Producer for parse and upload."""

import argparse
import asyncio
import hashlib
import logging
import time
from datetime import datetime, timedelta
from typing import Any

from aiobotocore.session import get_session
from botocore import UNSIGNED
from botocore.config import Config

logger = logging.getLogger(__name__)


def get_parser() -> argparse.ArgumentParser:
    """Get parser for command line arguments."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "bucket_name",
        type=str,
        help="Url of the provider where to download the articles.",
    )
    parser.add_argument(
        "queue_url",
        type=str,
        help="Url of the queue where to upload the articles.",
    )
    parser.add_argument(
        "--index",
        type=str,
        default="pmc_paragraphs",
        help="Desired name of the index holding paragraphs.",
    )
    parser.add_argument(
        "--parser-name",
        type=str,
        default=None,
        help=(
            "Endpoint of the parser to use. Specify only if the s3 bucket is not ours."
            " Else it will be fetch from the path."
        ),
    )
    parser.add_argument(
        "--start-date",
        type=str,
        default=None,
        help="Date of the oldest document to download. Format: dd-mm-yyy.",
    )
    parser.add_argument(
        "--prefixes",
        nargs="+",
        type=str,
        default=None,
        help=(
            "Prefix in the s3 path to filter the search on. Can write multiple prefixes"
            " to make it a list."
        ),
    )
    parser.add_argument(
        "--sign-request",
        action="store_true",
        help=(
            "Sign the request with AWS credential. Should be activated for our bucket,"
            " not for public external buckets."
        ),
    )
    parser.add_argument(
        "--file-extension",
        type=str,
        default=None,
        help="Extension of the file to enqueue. Leave None to parse any file type.",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Control verbosity",
    )
    return parser


async def run(
    bucket_name: str,
    queue_url: str,
    index: str,
    start_date: datetime,
    parser_name: str | None = None,
    prefixes: list[str | None] | None = None,
    sign_request: bool = False,
    file_extension: str | None = None,
) -> int:
    """Run the queue upload.

    Parameters
    ----------
    bucket_name
        Name of the s3 bucket where the articles are fetched.
    queue_url
        URL of the queue that holds the articles.
    index
        Index in the db where to send the parsed documents.
    start_date
        Date after which articles will be picked.
    parser_name
        Name of the parser to use to parse those documents. Leave None if using our private s3.
    prefixes
        Partial path inside the s3 bucket where to fetch articles.
    sign_request
        Sign the s3 request with AWS credentials. Should be True if there are documents from our private s3 bucket in the queue.
    file_extension
        Partial match of the name of the file (e.g. to specify a specific file type).

    Returns
    -------
    1 or 0
    """
    # Setup s3 bucket
    logger.info("Starting S3 client.")
    session = get_session()
    start_time = time.time()
    async with (
        session.create_client(
            "s3",
            region_name="us-east-1",
            config=Config(signature_version=UNSIGNED) if not sign_request else None,
        ) as s3_client,
        session.create_client("sqs", region_name="us-east-1") as q_client,
    ):
        prefixes = prefixes if prefixes else [None]
        for prefix in prefixes:
            s3_paginator = s3_client.get_paginator("list_objects_v2")
            s3_iterator = s3_paginator.paginate(
                Bucket=bucket_name, Prefix=prefix if prefix else ""
            )
            logger.info(
                f"Filtering interesting articles. Current prefix: {prefix}, current"
                f" date: {start_date}."
            )
            filtered_iterator = s3_iterator.search(
                f"""Contents[?to_string(LastModified)>='\"{start_date.strftime('%Y-%m-%d %H:%M:%S%')}+00:00\"'
                    && contains(Key, '.{file_extension if file_extension else ""}')]"""
            )
            finished = False
            while not finished:
                start_time_batch = time.time()

                messages = []
                i = 0
                parser_names = {
                    "jats_xml",
                    "tei_xml",
                    "xocs_xml",
                    "core_json",
                    "grobid_pdf",
                    "pypdf_pdf",
                    "pubmed_xml",
                }
                # Creating batches.
                async for obj in filtered_iterator:
                    i += 1
                    message: dict[str, Any] = {}
                    s3_bucket_key = obj["Key"]
                    message["Id"] = hashlib.md5(
                        s3_bucket_key.encode(), usedforsecurity=False
                    ).hexdigest()

                    # For our bucket, we assume a layout with 7 folders named after each parser, and corresponding papers inside.
                    # For external bucket, specify directly the name of the parser to use.
                    parser = (
                        parser_name
                        if parser_name
                        else next(
                            (term for term in parser_names if term in s3_bucket_key)
                        )
                    )
                    message["MessageBody"] = "Hello I am the producer."
                    message["MessageAttributes"] = {
                        "Key": {"DataType": "String", "StringValue": s3_bucket_key},
                        "Bucket_Name": {
                            "DataType": "String",
                            "StringValue": bucket_name,
                        },
                        "Parser_Endpoint": {
                            "DataType": "String",
                            "StringValue": parser,
                        },
                        # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs/client/send_message.html
                        "Sign": {
                            "DataType": "Number",
                            "StringValue": str(int(sign_request)),
                        },
                        "Index": {"DataType": "String", "StringValue": index},
                    }
                    messages.append(message)
                    # 10 is the maximum number of messages one can send per batch.
                    if i == 10:
                        break
                if i != 10:
                    finished = True
                if messages:
                    results = await q_client.send_message_batch(
                        QueueUrl=queue_url, Entries=messages
                    )
                else:
                    continue
                if "Failed" in results:
                    logger.info(f"Failed: {results['Failed']}")
                logger.info(
                    f"Done producing a batch. Took {time.time() - start_time_batch}s to"
                    f" process it. Total time: {time.time() - start_time}"
                )
    logger.info("Done.")

    return 0


def main() -> None:
    """Run the main logic."""
    # Parse command line
    parser = get_parser()
    args = parser.parse_args()

    # Additional parameters
    logging_level = logging.INFO if args.verbose else logging.WARNING

    # setup logging
    logging.basicConfig(
        format="[%(levelname)s]  %(asctime)s %(name)s  %(message)s", level=logging_level
    )
    if args.start_date is not None:
        start_date = datetime.strptime(args.start_date, "%d-%m-%Y")
        if start_date > datetime.today():
            raise ValueError(
                "Date is in the future! Hope you found the secret for time travelling."
            )
    else:
        start_date = datetime.today() - timedelta(days=1)
    args.__dict__["start_date"] = start_date
    args.__dict__.pop("verbose")
    raise SystemExit(asyncio.run(run(**args.__dict__)))


if __name__ == "__main__":
    main()
